#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest
from earthkit.utils.array import _CUPY
from earthkit.utils.array import _NUMPY
from earthkit.utils.array import _TORCH
from earthkit.utils.array import get_backend
from earthkit.utils.testing import NO_CUPY
from earthkit.utils.testing import NO_TORCH

from earthkit.data import FieldList
from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize("_kwargs", [{}, {"array_backend": "numpy"}])
def test_grib_file_numpy_backend(_kwargs):
    ds = from_source("file", earthkit_examples_file("test6.grib"))
    ds = ds.to_fieldlist(**_kwargs)

    assert getattr(ds, "path", None) is None
    assert len(ds) == 6

    assert ds[0].array_backend == _NUMPY
    assert isinstance(ds[0].values, np.ndarray)
    assert ds[0].values.shape == (84,)

    assert isinstance(ds.values, np.ndarray)
    assert ds.values.shape == (
        6,
        84,
    )

    assert isinstance(ds[0].to_array(), np.ndarray)
    assert ds[0].to_array().shape == (7, 12)

    assert isinstance(ds.to_array(), np.ndarray)
    assert ds.to_array().shape == (6, 7, 12)

    assert isinstance(ds[0].to_numpy(), np.ndarray)
    assert ds[0].to_numpy().shape == (7, 12)

    assert isinstance(ds.to_numpy(), np.ndarray)
    assert ds.to_numpy().shape == (6, 7, 12)

    assert get_backend(ds[0].to_array()) == _NUMPY

    ds1 = ds.to_fieldlist()
    assert len(ds1) == len(ds)
    assert getattr(ds1, "path", None) is None
    assert get_backend(ds1[0].to_array()) == _NUMPY
    assert ds1[0].array_backend == _NUMPY


@pytest.mark.skipif(NO_TORCH, reason="No pytorch installed")
def test_grib_file_pytorch_backend():
    ds = from_source("file", earthkit_examples_file("test6.grib"))
    ds = ds.to_fieldlist(array_backend="pytorch")

    assert getattr(ds, "path", None) is None
    assert len(ds) == 6

    import torch

    assert ds[0].array_backend == _TORCH
    assert torch.is_tensor(ds[0].values)
    assert ds[0].values.shape == (84,)

    assert torch.is_tensor(ds.values)
    assert ds.values.shape == (
        6,
        84,
    )

    x = ds[0].to_array()
    assert torch.is_tensor(x)
    assert x.shape == (7, 12)

    x = ds.to_array()
    assert torch.is_tensor(x)
    assert x.shape == (6, 7, 12)

    x = ds[0].to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (7, 12)

    x = ds.to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (6, 7, 12)

    assert get_backend(ds[0].to_array()) == _TORCH

    ds1 = ds.to_fieldlist()
    assert len(ds1) == len(ds)
    assert getattr(ds1, "path", None) is None
    assert get_backend(ds1[0].to_array()) == _TORCH
    assert ds1[0].array_backend == _TORCH


@pytest.mark.skipif(NO_CUPY, reason="No cupy installed")
def test_grib_file_cupy_backend():
    ds = from_source("file", earthkit_examples_file("test6.grib"))
    ds = ds.to_fieldlist(array_backend="cupy")

    import cupy as cp

    assert getattr(ds, "path", None) is None
    assert len(ds) == 6

    assert isinstance(ds[0].values, cp.ndarray)
    assert ds[0].values.shape == (84,)

    assert isinstance(ds.values, cp.ndarray)
    assert ds.values.shape == (
        6,
        84,
    )

    x = ds[0].to_array()
    assert isinstance(x, cp.ndarray)
    assert x.shape == (7, 12)

    x = ds.to_array()
    assert isinstance(x, cp.ndarray)
    assert x.shape == (6, 7, 12)

    x = ds[0].to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (7, 12)

    x = ds.to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (6, 7, 12)

    assert get_backend(ds[0].to_array()) == _CUPY

    ds1 = ds.to_fieldlist()
    assert len(ds1) == len(ds)
    assert getattr(ds1, "path", None) is None
    assert get_backend(ds1[0].to_array()) == _CUPY


def test_grib_array_numpy_backend():
    s = from_source("file", earthkit_examples_file("test6.grib"))

    ds = FieldList.from_array(
        s.values,
        [m for m in s.metadata()],
    )
    assert len(ds) == 6
    with pytest.raises(AttributeError):
        ds.path

    assert isinstance(ds[0].values, np.ndarray)
    assert ds[0].values.shape == (84,)

    assert isinstance(ds.values, np.ndarray)
    assert ds.values.shape == (
        6,
        84,
    )

    assert isinstance(ds[0].to_array(), np.ndarray)
    assert ds[0].to_array().shape == (7, 12)

    assert isinstance(ds.to_array(), np.ndarray)
    assert ds.to_array().shape == (6, 7, 12)

    assert isinstance(ds[0].to_numpy(), np.ndarray)
    assert ds[0].to_numpy().shape == (7, 12)

    assert isinstance(ds.to_numpy(), np.ndarray)
    assert ds.to_numpy().shape == (6, 7, 12)


@pytest.mark.skipif(NO_TORCH, reason="No pytorch installed")
def test_grib_array_torch_backend():
    s = from_source("file", earthkit_examples_file("test6.grib"))

    import torch

    ds = FieldList.from_array(
        torch.tensor(s.values),
        [m for m in s.metadata()],
    )
    assert len(ds) == 6
    with pytest.raises(AttributeError):
        ds.path

    assert torch.is_tensor(ds[0].values)
    assert ds[0].values.shape == (84,)

    assert torch.is_tensor(ds.values)
    assert ds.values.shape == (
        6,
        84,
    )

    assert torch.is_tensor(ds[0].to_array())
    assert ds[0].to_array().shape == (7, 12)

    assert torch.is_tensor(ds.to_array())
    assert ds.to_array().shape == (6, 7, 12)

    assert isinstance(ds[0].to_numpy(), np.ndarray)
    assert ds[0].to_numpy().shape == (7, 12)

    assert isinstance(ds.to_numpy(), np.ndarray)
    assert ds.to_numpy().shape == (6, 7, 12)


@pytest.mark.skipif(NO_CUPY, reason="No cupy installed")
def test_grib_array_cupy_backend():
    s = from_source("file", earthkit_examples_file("test6.grib"))

    import cupy as cp

    ds = FieldList.from_array(
        cp.array(s.values),
        [m for m in s.metadata()],
    )
    assert len(ds) == 6
    with pytest.raises(AttributeError):
        ds.path

    assert isinstance(ds[0].values, cp.ndarray)
    assert ds[0].values.shape == (84,)

    assert isinstance(ds.values, cp.ndarray)
    assert ds.values.shape == (
        6,
        84,
    )

    x = ds[0].to_array()
    assert isinstance(x, cp.ndarray)
    assert x.shape == (7, 12)

    x = ds.to_array()
    assert isinstance(x, cp.ndarray)
    assert x.shape == (6, 7, 12)

    x = ds[0].to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (7, 12)

    x = ds.to_numpy()
    assert isinstance(x, np.ndarray)
    assert x.shape == (6, 7, 12)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import ArrayField
from earthkit.data import SimpleFieldList
from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def _check(ds, group):
    assert len(ds) == 6

    ref = [("t", 1000), ("u", 1000), ("v", 1000), ("t", 850), ("u", 850), ("v", 850)]

    assert ds.metadata(("param", "level")) == ref

    ref = [
        [("t", 1000), ("t", 850)],
        [("u", 1000), ("u", 850)],  #
        [("v", 1000), ("v", 850)],
    ]
    cnt = 0
    for i, f in enumerate(ds.group_by(group)):
        assert len(f) == 2
        assert f.metadata(("param", "level")) == ref[i]
        afl = f.to_fieldlist()
        assert afl is not f
        assert len(afl) == 2
        cnt += len(f)

    assert cnt == len(ds)


@pytest.mark.parametrize("group", ["param"])
def test_grib_simple_fl_1(group):
    ds_in = from_source("file", earthkit_examples_file("test6.grib"))

    ds = SimpleFieldList()
    for f in ds_in:
        ds.append(f)

    _check(ds, group)


@pytest.mark.parametrize("group", ["param"])
def test_grib_simple_fl_2(group):
    ds = from_source("file", earthkit_examples_file("test6.grib"))

    ds = SimpleFieldList([f for f in ds])

    _check(ds, group)


@pytest.mark.parametrize("group", ["param"])
def test_grib_simple_fl_3(group):
    ds_in = from_source("file", earthkit_examples_file("test6.grib"))

    ds = SimpleFieldList()
    for f in ds_in:
        ds.append(ArrayField(f.to_numpy(), f.metadata()))

    _check(ds, group)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.sources.multi_url import MultiUrl
from earthkit.data.sources.url import Url
from earthkit.data.utils import ensure_sequence

from .file import FileSource

LOG = logging.getLogger(__name__)

DEFAULT_ENDPOINT = "s3.amazonaws.com"
DEFAULT_REGION = "eu-west-2"


def request_to_resource(requests, global_credentials):
    resources = []
    for r in requests:
        bucket = r["bucket"]
        endpoint = r.get("endpoint", DEFAULT_ENDPOINT)
        o = ensure_sequence(r["objects"])
        credentials = S3Credentials.from_other(global_credentials)
        credentials.update(r)
        credentials.check()
        for obj in o:
            if isinstance(obj, str):
                key = obj
                resources.append(S3Resource(endpoint, bucket, DEFAULT_REGION, key))
            else:
                key = obj["object"]
                region = obj.get("region", DEFAULT_REGION)
                parts = obj.get("parts", None)
                resources.append(
                    S3Resource(
                        endpoint,
                        bucket,
                        region,
                        key,
                        parts=parts,
                        credentials=credentials,
                    )
                )

    return resources


class S3Credentials:
    def __init__(self, aws_access_key, aws_secret_access_key, aws_token=None):
        self.aws_access_key = aws_access_key
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_token = aws_token

    def check(self):
        if self.aws_access_key and not self.aws_secret_access_key:
            raise ValueError("When aws_access_key is provided, aws_secret_access_key must also be provided")

        if not self.aws_access_key and self.aws_secret_access_key:
            raise ValueError("When aws_secret_access_key is provided, aws_access_key must also be provided")

    @classmethod
    def from_other(self, other):
        if other is None:
            return self(None, None)
        return self(*[other.aws_access_key, other.aws_secret_access_key, other.aws_token])

    def valid(self):
        return self.aws_access_key and self.aws_secret_access_key

    def update(self, obj):
        self.aws_access_key = self.aws_access_key or obj.get("aws_access_key", None)
        self.aws_secret_access_key = self.aws_secret_access_key or obj.get("aws_secret_access_key", None)
        self.aws_token = self.aws_token or obj.get("aws_token", None)


class S3Authenticator:
    def __init__(self, region, credentials=None):
        self.region = region
        self.credentials = credentials

    @staticmethod
    def _host(url):
        from urllib.parse import urlparse

        return urlparse(url).netloc

    def __call__(self, r):
        if self.credentials is not None and self.credentials.valid():
            return self._base_auth(r)

        return self._boto_auth(r)

    def _base_auth(self, r):
        assert self.credentials is not None
        assert self.credentials.valid()
        from aws_requests_auth import AWSRequestsAuth

        host = self._host(r.url)
        auth = AWSRequestsAuth(
            aws_access_key=self.credentials.aws_access_key,
            aws_secret_access_key=self.credentials.aws_secret_access_key,
            aws_host=host,
            aws_region=self.region,
            aws_service="s3",
            aws_token=self.credentials.aws_token,
        )
        return auth(r)

    def _boto_auth(self, r):
        from aws_requests_auth.boto_utils import BotoAWSRequestsAuth

        host = self._host(r.url)
        auth = BotoAWSRequestsAuth(
            aws_host=host,
            aws_region=self.region,
            aws_service="s3",
        )

        # BotoAWSRequestsAuth raises AttributeError when no credentials are found
        # due to calling a method on a None object. Until this is handled
        # correctly in aws_requests_auth we raise a properly worded exception.
        try:
            return auth(r)
        except AttributeError:
            raise Exception(
                (
                    "No S3 credentials were found using botocore. See the following page "
                    "about how credentials are searched for: http://boto3.readthedocs.io/en/"
                    "latest/guide/configuration.html#configuring-credentials"
                )
            ) from None


class S3Resource:
    def __init__(self, endpoint, bucket, region, key, parts=None, credentials=None):
        self.endpoint = endpoint
        self.region = region
        self.bucket = bucket
        self.key = key
        self.parts = parts
        self.credentials = credentials

    @property
    def url(self):
        if "amazonaws" in self.endpoint:
            return f"https://{self.bucket}.{self.endpoint}/{self.key}"
        else:
            return f"https://{self.endpoint}/{self.bucket}/{self.key}"

    def host(self):
        if "amazonaws" in self.endpoint:
            return f"{self.bucket}.{self.endpoint}"
        else:
            return f"{self.endpoint}"

    def __repr__(self):
        return f"{self.__class__.__name__}(url={self.url}, part={self.part})"


class S3Source(FileSource):
    """Represent an AWS S3 bucket source"""

    def __init__(
        self,
        *args,
        stream=False,
        read_all=False,
        anon=True,
        aws_access_key=None,
        aws_secret_access_key=None,
        aws_token=None,
        **kwargs,
    ) -> None:
        super().__init__()

        self.anon = anon

        self.stream = stream
        self._stream_kwargs = dict(read_all=read_all)

        def _collect(r):
            if isinstance(r, dict):
                self.request.append(r)
            elif isinstance(r, (list, tuple)):
                for x in r:
                    _collect(x)
            else:
                raise ValueError(f"Invalid request: {r}")

        self.request = []
        for a in args:
            _collect(a)

        self.credentials = S3Credentials(aws_access_key, aws_secret_access_key, aws_token)
        self.resources = request_to_resource(self.request, self.credentials)

    def mutate(self):
        url_spec = []
        has_parts = any(r.parts is not None for r in self.resources)
        for r in self.resources:
            spec = {"url": r.url}
            if has_parts:
                spec["parts"] = r.parts
            if not self.anon:
                auth = S3Authenticator(r.region, credentials=r.credentials)
                spec["auth"] = auth
            url_spec.append(spec)

        if not self.anon and has_parts:
            fake_headers = {"accept-ranges": "bytes"}
        else:
            fake_headers = None

        if self.stream:
            return Url(
                url_spec,
                fake_headers=fake_headers,
                stream=True,
                **self._stream_kwargs,
            )
        else:
            return MultiUrl(url_spec, sort_urls=False, fake_headers=fake_headers)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


source = S3Source



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np
import pytest
import xarray as xr

from earthkit.data import from_source
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.testing import load_nc_or_xr_source

# These functionalities are variations around
# http://xarray.pydata.org/en/stable/user-guide/combining.html#combining-multi


def assert_same_xarray(x, y):
    assert x.broadcast_equals(y)
    assert x.equals(y)
    assert x.identical(y)
    assert len(x) == len(y)
    assert set(x.keys()) == set(y.keys())
    assert len(x.dims) == len(y.dims)
    assert len(x.coords) == len(y.coords)
    for k in x.keys():
        xda, yda = x[k], y[k]
        assert xda.values.shape == yda.values.shape
        assert np.all(xda.values == yda.values)


def merger_func(paths_or_sources):
    return xr.open_mfdataset(paths_or_sources)


class Merger_obj:
    def to_xarray(self, paths_or_sources, **kwargs):
        return xr.open_mfdataset(paths_or_sources)


@pytest.mark.parametrize("mode", ["nc", "xr"])
def test_netcdf_concat(mode):
    ds1 = load_nc_or_xr_source(earthkit_test_data_file("era5_2t_1.nc"), mode)
    ds2 = load_nc_or_xr_source(earthkit_test_data_file("era5_2t_2.nc"), mode)
    ds = ds1 + ds2

    assert len(ds) == 2
    md = ds1.metadata("variable") + ds2.metadata("variable")
    assert ds.metadata("variable") == md

    assert ds[0].datetime() == {
        "base_time": datetime.datetime(2021, 3, 1, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 1, 12, 0),
    }
    assert ds[1].datetime() == {
        "base_time": datetime.datetime(2021, 3, 2, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 2, 12, 0),
    }
    assert ds.datetime() == {
        "base_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
    }

    import xarray as xr

    target = xr.merge([ds1.to_xarray(), ds2.to_xarray()])
    merged = ds.to_xarray()
    assert target.identical(merged)


def test_netcdf_read_multiple_files():
    ds = from_source(
        "file",
        [
            earthkit_test_data_file("era5_2t_1.nc"),
            earthkit_test_data_file("era5_2t_2.nc"),
        ],
    )

    assert len(ds) == 2
    assert ds.metadata("variable") == ["t2m", "t2m"]

    assert ds[0].datetime() == {
        "base_time": datetime.datetime(2021, 3, 1, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 1, 12, 0),
    }
    assert ds[1].datetime() == {
        "base_time": datetime.datetime(2021, 3, 2, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 2, 12, 0),
    }
    assert ds.datetime() == {
        "base_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
    }

    import xarray as xr

    target = xr.merge(
        [
            xr.open_dataset(earthkit_test_data_file("era5_2t_1.nc")),
            xr.open_dataset(earthkit_test_data_file("era5_2t_2.nc")),
        ]
    )
    merged = ds.to_xarray()
    assert target.identical(merged)


@pytest.mark.parametrize("custom_merger", (merger_func, Merger_obj()))
def test_netdcf_merge_custom(custom_merger):
    s1 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["a", "b"],
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["c", "d"],
    )
    ds2 = s2.to_xarray()

    target = xr.merge([ds1, ds2])

    ds = from_source("multi", [s1, s2], merger=custom_merger)
    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged)

    target2 = xr.open_mfdataset([s1.path, s2.path])
    assert target2.identical(merged)


def test_netcdf_merge_var():
    s1 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["a", "b"],
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["c", "d"],
    )
    ds2 = s2.to_xarray()

    target = xr.merge([ds1, ds2])
    ds = from_source("multi", [s1, s2])
    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged)

    target2 = xr.open_mfdataset([s1.path, s2.path])
    assert target2.identical(merged)


def _merge_var_different_coords(kind1, kind2):
    s1 = from_source(
        "dummy-source",
        kind=kind1,
        dims=["lat", "lon"],
        variables=["a", "b"],
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind=kind2,
        dims=["lat", "time"],
        variables=["c", "d"],
    )
    ds2 = s2.to_xarray()

    target = xr.merge([ds1, ds2])
    ds = from_source("multi", [s1, s2])
    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged)


def test_netcdf_merge_var_different_coords():
    _merge_var_different_coords("netcdf", "netcdf")


@pytest.mark.skipif(True, reason="Test not yet implemented")
def test_grib_merge_var_different_coords():
    _merge_var_different_coords("grib", "grib")


@pytest.mark.skipif(True, reason="Test not yet implemented")
def test_grib_nc_merge_var_different_coords():
    _merge_var_different_coords("netcdf", "grib")


def _concat_var_different_coords_1(kind1, kind2):
    s1 = from_source(
        "dummy-source",
        kind=kind1,
        variables=["a"],
        dims=["lat", "lon", "time"],
        coord_values=dict(time=[1, 3]),
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind=kind2,
        variables=["a"],
        dims=["lat", "lon", "time"],
        coord_values=dict(time=[2, 4]),
    )
    ds2 = s2.to_xarray()

    target = xr.concat([ds1, ds2], dim="time")

    ds = from_source("multi", [s1, s2], merger="concat(concat_dim=time)")
    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged), f"Concat failed for {kind1}, {kind2}"


def test_netcdf_concat_var_different_coords_1():
    for kind1 in ["netcdf"]:  # ["netcdf", "grib"]:
        for kind2 in ["netcdf"]:  # ["netcdf", "grib"]:
            _concat_var_different_coords_1(kind1, kind2)


def test_netcdf_concat_var_different_coords_2():
    s1 = from_source(
        "dummy-source",
        kind="netcdf",
        variables=["a"],
        dims=["lat", "lon", "time"],
        coord_values=dict(time=[2, 1]),
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind="netcdf",
        variables=["a"],
        dims=["lat", "lon", "time"],
        coord_values=dict(time=[3, 4]),
    )
    ds2 = s2.to_xarray()

    target = xr.concat([ds1, ds2], dim="time")

    ds = from_source("multi", [s1, s2], merger="concat(concat_dim=time)")

    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged)


def test_netcdf_wrong_concat_var():
    s1 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["a", "b"],
        coord_values=dict(time=[1, 2]),
    )
    ds1 = s1.to_xarray()

    s2 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "time"],
        variables=["a", "b"],
        coord_values=dict(time=[8, 9]),
    )
    ds2 = s2.to_xarray()

    print(f"s1={s1}")
    print(f"s2={s2}")
    target = xr.concat([ds1, ds2], dim="time")
    ds = from_source("multi", [s1, s2], merger="concat(concat_dim=time)")

    ds.graph()
    merged = ds.to_xarray()

    assert target.identical(merged)


def get_hierarchy():
    a1 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "forecast_time"],
        variables=["a"],
        coord_values=dict(forecast_time=[1, 3]),
    )
    a2 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "forecast_time"],
        variables=["a"],
        coord_values=dict(forecast_time=[2, 4]),
    )
    b1 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "forecast_time"],
        variables=["b"],
        coord_values=dict(forecast_time=[1, 3]),
    )
    b2 = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "forecast_time"],
        variables=["b"],
        coord_values=dict(forecast_time=[2, 4]),
    )

    target = xr.merge(
        [
            xr.merge([a1.to_xarray(), a2.to_xarray()]),
            xr.merge([b1.to_xarray(), b2.to_xarray()]),
        ]
    )
    return target, a1, a2, b1, b2


@pytest.mark.skipif(True, reason="Test not yet implemented")
def test_nc_concat_merge_var():
    target, a1, a2, b1, b2 = get_hierarchy()

    s = from_source(
        "multi",
        [
            from_source("multi", [a1, a2], merger="concat(dim=forecast_time)"),
            from_source("multi", [b1, b2], merger="concat(dim=forecast_time)"),
        ],
        merger="merge",
    )

    merged = s.to_xarray()
    assert target.identical(merged), merged


@pytest.mark.skipif(True, reason="Test not yet implemented")
def test_netcdf_merge_concat_var():
    target, a1, a2, b1, b2 = get_hierarchy()
    s = from_source(
        "multi",
        [
            from_source("multi", [a1, b1], merger="merge()"),
            from_source("multi", [a2, b2], merger="merge()"),
        ],
        merger="concat(dim=forecast_time)",
    )
    merged = s.to_xarray()
    assert target.identical(merged)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from .. import Reader
from .fieldlist import NetCDFFieldListFromFile
from .fieldlist import NetCDFFieldListFromURL


class NetCDFFieldListReader(NetCDFFieldListFromFile, Reader):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        NetCDFFieldListFromFile.__init__(self, path)

    def __repr__(self):
        return "NetCDFFieldListReader(%s)" % (self.path,)

    def mutate_source(self):
        # A NetCDFReader is a source itself
        return self


class NetCDFFieldListUrlReader(NetCDFFieldListFromURL, Reader):
    def __init__(self, source, url):
        Reader.__init__(self, source, url)
        NetCDFFieldListFromURL.__init__(self, url)

    def __repr__(self):
        return "NetCDFFieldListUrlReader(%s)" % (self.path,)

    # def mutate_source(self):
    #     # A NetCDFReader is a source itself
    #     return self


class NetCDFReader(Reader):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)

    def __repr__(self):
        return "NetCDFReader(%s)" % (self.path,)

    def to_numpy(self, flatten=False):
        arr = self.to_xarray().to_array().to_numpy()
        if flatten:
            arr = arr.flatten()
        return arr

    def to_pandas(self):
        return self.to_xarray().to_pandas()

    def to_xarray(self, **kwargs):
        return type(self).to_xarray_multi_from_paths([self.path], **kwargs)

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        if not isinstance(paths, list):
            paths = [paths]

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))
        if not options:
            options = dict(**kwargs)

        return xr.open_mfdataset(
            paths,
            **options,
        )


def _match_magic(magic, deeper_check):
    if magic is not None:
        type_id = (b"\x89HDF", b"CDF\x01", b"CDF\x02")
        return len(magic) >= 4 and magic[:4] in type_id
    return False


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    if _match_magic(magic, deeper_check):
        fs = NetCDFFieldListReader(source, path)
        if fs.has_fields():
            return fs
        else:
            return NetCDFReader(source, path)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import subprocess

from .file import FileSource

LOG = logging.getLogger(__name__)


class ECFSRetriever(FileSource):
    def __init__(
        self,
        url,
        *,
        force=False,
        **kwargs,
    ):
        super().__init__(url, **kwargs)

        self.path = self.cache_file(
            self._fetch,
            url,
            force=force,
        )

    def _fetch(self, target, url):
        if not url.startswith("ec:"):
            url = f"ec:{url}"

        LOG.debug("%s", f"Fetching {url} to {target}")

        subprocess.run(["ecp", url, target], check=True)


source = ECFSRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import deprecation

from earthkit.data.readers import memory_reader

from . import Source

LOG = logging.getLogger(__name__)


class MemoryBaseSource(Source):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __iter__(self):
        return iter(self._reader)

    def __len__(self):
        return len(self._reader)

    def __getitem__(self, n):
        return self._reader[n]

    def sel(self, *args, **kwargs):
        return self._reader.sel(*args, **kwargs)

    def order_by(self, *args, **kwargs):
        return self._reader.order_by(*args, **kwargs)

    def to_xarray(self, **kwargs):
        return self._reader.to_xarray(**kwargs)

    def to_pandas(self, **kwargs):
        return self._reader.to_pandas(**kwargs)

    def to_numpy(self, **kwargs):
        return self._reader.to_numpy(**kwargs)

    @property
    def values(self):
        return self._reader.values

    def to_fieldlist(self, *args, **kwargs):
        return self._reader.to_fieldlist(*args, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def save(self, path, **kwargs):
        return self.to_target("file", path, **kwargs)
        # return self._reader.save(path, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def write(self, f, **kwargs):
        return self.to_target("file", f, **kwargs)
        # return self._reader.write(f, **kwargs)

    def to_target(self, *args, **kwargs):
        return self._reader.to_target(*args, **kwargs)

    def scaled(self, *args, **kwargs):
        return self._reader.scaled(*args, **kwargs)

    def _attributes(self, names):
        return self._reader._attributes(names)

    # def __fspath__(self):
    #     return self.path

    def metadata(self, *args, **kwargs):
        return self._reader.metadata(*args, **kwargs)

    def ls(self, *args, **kwargs):
        return self._reader.ls(*args, **kwargs)

    def describe(self, *args, **kwargs):
        return self._reader.describe(*args, **kwargs)

    def datetime(self, **kwargs):
        return self._reader.datetime(**kwargs)

    def bounding_box(self):
        return self._reader.bounding_box()

    def statistics(self, **kwargs):
        return self._reader.statistics(**kwargs)

    def batched(self, *args):
        return self._reader.batched(*args)

    def group_by(self, *args):
        return self._reader.group_by(*args)


class MemorySource(MemoryBaseSource):
    def __init__(self, buf, **kwargs):
        self._buf = buf
        self._reader_ = None

    def mutate(self):
        source = self._reader.mutate_source()
        if source not in (None, self):
            source._parent = self
            return source
        return self

    @property
    def _reader(self):
        if self._reader_ is None:
            self._reader_ = memory_reader(self, self._buf)
            self._buf = None
        return self._reader_


source = MemorySource



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import NO_FDB
from earthkit.data.testing import NO_PROD_FDB
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.skipif(NO_PROD_FDB, reason="No access to operational FDB")
def test_fdb_grib_stream():
    d = datetime.datetime.now() - datetime.timedelta(days=2)
    request = {
        "class": "od",
        "expver": "0001",
        "stream": "oper",
        "date": d.strftime("%Y%m%d"),
        "time": [0, 12],
        "domain": "g",
        "type": "an",
        "levtype": "sfc",
        "step": 0,
        "param": [151, 167],
    }

    ds = from_source("fdb", request)
    cnt = sum([1 for f in ds])
    assert cnt == 4


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.skipif(NO_PROD_FDB, reason="No access to operational FDB")
def test_fdb_grib_file():
    d = datetime.datetime.now() - datetime.timedelta(days=2)
    request = {
        "class": "od",
        "expver": "0001",
        "stream": "oper",
        "date": d.strftime("%Y%m%d"),
        "time": [0, 12],
        "domain": "g",
        "type": "an",
        "levtype": "sfc",
        "step": 0,
        "param": [151, 167],
    }

    ds = from_source("fdb", request, stream=False)
    assert len(ds) == 4


def make_fdb_config(path):
    fdb_schema = earthkit_test_data_file("fdb_schema.txt")
    fdb_dir = path
    os.makedirs(fdb_dir, exist_ok=True)
    config = {
        "type": "local",
        "engine": "toc",
        "schema": fdb_schema,
        "spaces": [{"handler": "Default", "roots": [{"path": fdb_dir}]}],
    }
    return config


@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
@pytest.mark.parametrize("use_kwargs", [True, False])
def test_fdb_grib_write(monkeypatch, use_kwargs):
    with temp_directory() as tmpdir:
        config = make_fdb_config(os.path.join(tmpdir, "_fdb"))

        ds = from_source("file", earthkit_examples_file("test.grib"))

        import pyfdb

        fdb = pyfdb.FDB(config=config)
        for f in ds:
            fdb.archive(f.message())
        fdb.flush()

        request = {
            "class": "od",
            "expver": "0001",
            "stream": "oper",
            "date": "20200513",
            "time": 1200,
            "domain": "g",
            "type": "an",
            "levtype": "sfc",
            "step": 0,
            "param": [167, 151],
        }

        if use_kwargs:
            ds1 = from_source("fdb", request, config=config, stream=False)
        else:
            monkeypatch.setenv("FDB5_CONFIG", str(config))
            ds1 = from_source("fdb", request, stream=False)

        assert len(ds1) == 2
        assert ds1.metadata("param") == ["2t", "msl"]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
with open("res.grib", "wb") as f:
    ds.write(f)

# case 2
with open("res.grib", "wb") as f:
    for field in ds:
        field.write(f)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import logging
from functools import cached_property

from earthkit.data.core.fieldlist import FieldList
from earthkit.data.readers import stream_reader
from earthkit.data.sources.memory import MemoryBaseSource

from . import Source

LOG = logging.getLogger(__name__)


def parse_stream_kwargs(**kwargs):
    read_all = kwargs.pop("read_all", False)
    stream_kwargs = dict(read_all=read_all)
    return (stream_kwargs, kwargs)


class StreamMemorySource(MemoryBaseSource):
    def __init__(self, stream, **kwargs):
        super().__init__(**kwargs)
        if not isinstance(stream, Stream):
            raise ValueError(f"Invalid stream={stream}")
        self._stream = stream

    @cached_property
    def _reader(self):
        reader = stream_reader(self, self._stream.stream, True, **self._kwargs)
        if reader is None:
            raise TypeError(f"could not create reader for stream={self._stream}")
        return reader

    def mutate(self):
        source = self._reader.mutate_source()
        if source not in (None, self):
            # source._parent = self
            return source
        return self


class StreamSource(Source):
    def __init__(self, stream, *, read_all=False, **kwargs):
        super().__init__()
        self._stream = self._wrap_stream(stream)
        self.memory = read_all

        # TODO: remove this check in a future release
        for k in ["group_by", "batch_size"]:
            if k in kwargs:
                raise ValueError(f"Invalid argument '{k}' for StreamSource. Deprecated since 0.8.0.")

        self._kwargs = kwargs

    def __iter__(self):
        return iter(self._reader)

    def mutate(self):
        if isinstance(self._stream, (list, tuple)):
            return MultiStreamSource(self._stream, read_all=self.memory)
        else:
            if self.memory:
                return StreamMemorySource(self._stream, **self._kwargs)
            elif hasattr(self._reader, "to_fieldlist"):
                return StreamFieldList(self._reader, **self._kwargs)
        return self

    @cached_property
    def _reader(self):
        reader = stream_reader(self, self._stream.stream, False, **self._kwargs)
        if reader is None:
            raise TypeError(f"could not create reader for stream={self._stream.stream}")
        return reader

    def batched(self, n):
        """Iterate through the stream in batches of ``n``.

        Parameters
        ----------
        n: int
            Batch size.

        Returns
        -------
        object
            Returns an iterator yielding batches of ``n`` elements. Each batch is a new object
            containing a view to the data in the original object, so no data is copied. The last
            batch may contain fewer than ``n`` elements.

        """
        return self._reader.batched(n)

    def group_by(self, *keys, **kwargs):
        """Iterate through the stream in groups defined by metadata keys.

        Parameters
        ----------
        *keys: tuple
            Positional arguments specifying the metadata keys to group by.
            Keys can be a single or multiple str, or a list or tuple of str.

        Returns
        -------
        object
            Returns an iterator yielding batches of elements grouped by the metadata ``keys``. Each
            batch is a new object containing a view to the data in the original object, so no data
            is copied. It generates a new group every time the value of the ``keys`` change.

        """
        return self._reader.group_by(*keys)

    def _wrap_stream(self, stream):
        if isinstance(stream, (list, tuple)):
            r = []
            for s in stream:
                if not isinstance(s, Stream):
                    r.append(Stream(s))
                else:
                    r.append(s)
            return r
        elif not isinstance(stream, Stream):
            return Stream(stream)

        return stream


class MultiStreamSource(Source):
    def __init__(self, sources, read_all=False, **kwargs):
        super().__init__(**kwargs)
        self.memory = read_all
        sources = self._from_sources(sources)
        self.sources = [s.mutate() for s in sources if not s.ignore()]

    def mutate(self):
        if self.memory:
            from .multi import MultiSource

            return MultiSource([s.mutate() for s in self.sources])
        else:
            if not any(isinstance(s, StreamFieldList) for s in self.sources):
                first = self.sources[0]
                if hasattr(first._reader, "to_fieldlist"):
                    return StreamFieldList(self, **self._kwargs)
            if all(isinstance(s, StreamFieldList) for s in self.sources):
                return StreamFieldList(self, **self._kwargs)

        return self

    def __iter__(self):
        return itertools.chain(*self.sources)

    def batched(self, n):
        from earthkit.data.utils.batch import batched

        return batched(self, n)

    def group_by(self, *args):
        from earthkit.data.utils.batch import group_by

        return group_by(self, *args)

    def _from_sources(self, sources):
        r = []
        for s in sources:
            if isinstance(s, (StreamSource, StreamFieldList, StreamMemorySource)):
                r.append(s)
            elif isinstance(s, Stream):
                r.append(StreamSource(s, read_all=self.memory))
            elif isinstance(s, MultiStreamSource):
                r.extend(s.sources)
            else:
                raise TypeError(f"Invalid source={s}")
        return r

    def to_xarray(self, **kwargs):
        from earthkit.data.core.fieldlist import FieldList

        fields = [f for f in self]
        return FieldList.from_fields(fields).to_xarray(**kwargs)

    @classmethod
    def merge(cls, sources):
        s = []
        for source in sources:
            if isinstance(source, StreamSource):
                s.append(source)
            elif isinstance(source, MultiStreamSource):
                s.extend(source.sources)
            else:
                s.append(source)

        return MultiStreamSource(s)


class StreamFieldList(FieldList, Source):
    def __init__(self, source, **kwargs):
        FieldList.__init__(self, **kwargs)
        self._source = source

    def mutate(self):
        return self

    def __iter__(self):
        return iter(self._source)

    def batched(self, n):
        return self._source.batched(n)

    def group_by(self, *keys, **kwargs):
        return self._source.group_by(*keys)

    def __getstate__(self):
        raise NotImplementedError("StreamFieldList cannot be pickled")

    def to_xarray(self, **kwargs):
        from earthkit.data.core.fieldlist import FieldList

        fields = [f for f in self]
        return FieldList.from_fields(fields).to_xarray(**kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(s, StreamFieldList) for s in sources), sources
        assert len(sources) > 1
        return MultiStreamSource.merge(sources)

    def default_encoder(self):
        return None


class Stream:
    def __init__(self, stream=None, maker=None, **kwargs):
        self._stream = stream
        self.maker = maker
        self.kwargs = kwargs
        if self._stream is None and self.maker is None:
            raise ValueError("Either stream or maker must be provided")

    @property
    def stream(self):
        if self._stream is None:
            self._stream = self.maker()
        return self._stream

    def ignore(self):
        return False

    def mutate(self):
        return self


def make_stream_source_from_other(source, **kwargs):
    stream_kwargs, kwargs = parse_stream_kwargs(**kwargs)

    if not isinstance(source, (list, tuple)):
        source = [source]

    for i, s in enumerate(source):
        stream = Stream(maker=s.to_stream)
        source[i] = StreamSource(stream, **stream_kwargs, **kwargs)

    if len(source) == 1:
        return source[0]
    else:
        return MultiStreamSource(source, **stream_kwargs)


source = StreamSource



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from io import IOBase

from . import SimpleTarget

LOG = logging.getLogger(__name__)


class FileTarget(SimpleTarget):
    """
    File target.

    Parameters:
    -----------
    file: str, file-like, None
        The file path or file-like object to write to. When None, tries to guess the file name
        from the ``data`` if it is passed as a kwarg.
        When the file name cannot be constructed, a ValueError is raised.
        When ``file`` is a path, a file object is automatically created and closed when the target is closed.
        When ``file`` is a file object, its ownership is not transferred to the target. As a consequence,
        the file object is not closed when the target is closed, even if :obj:`close` is called explicitly.
    append: bool
        If True, the file is opened in append mode. Only used if ``file`` is a path.
    **kwargs:
        Additional keyword arguments passed to the parent class.

    Raises:
    -------
    ValueError: If the file name is not specified and cannot be constructed.
    """

    def __init__(self, file=None, *, append=False, **kwargs):
        super().__init__(**kwargs)

        self.fileobj = None
        self._tmp_fileobj = None
        self.filename = None
        self.append = append
        self.ext = None

        if isinstance(file, IOBase):
            self.fileobj = file
        else:
            self.filename = file
            if self.filename is not None:
                _, self.ext = os.path.splitext(self.filename)

            if self.filename is None:
                self.filename = self._guess_filename(*kwargs)

            if not self.filename:
                raise ValueError("Please provide an output filename")

    def close(self):
        """Close the file if :obj:`FileTarget` was created with a file path.

        If :obj:`FileTarget` was created with a file object this call has no effect.
        The target will not be able to write anymore.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        self._mark_closed()
        if self._tmp_fileobj:
            self._tmp_fileobj.close()

    def flush(self):
        """Flush the file.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        self._f().flush()

    def _f(self):
        self._raise_if_closed()

        if self.fileobj:
            return self.fileobj

        if not self._tmp_fileobj:
            flag = "wb" if not self.append else "ab"
            self._tmp_fileobj = open(self.filename, flag)
        return self._tmp_fileobj

    def _guess_filename(self, data=None):
        """Try to guess filename from data when not provided"""
        if data is not None:
            for attr in ["source_filename", "path"]:
                if hasattr(self, attr) and getattr(self, attr) is not None:
                    return [os.path.basename(getattr(self, attr))]

    def _check_overwrite(self, data):
        """Ensure we do not overwrite file that is being read"""
        if (
            data is not None
            and self.filename is not None
            and os.path.isfile(self.filename)
            and hasattr(data, "path")
            and data.path is not None
            and os.path.isfile(data.path)
            and os.path.samefile(self.filename, data.path)
        ):
            import warnings

            warnings.warn(UserWarning(f"Earthkit refusing to overwrite the file being read: {self.filename}"))
            return False
        return True

    def _write(self, data=None, **kwargs):
        if not self._check_overwrite(data):
            return

        r = self._encode(data, suffix=self.ext, **kwargs)
        if hasattr(r, "__iter__"):
            f = self._f()
            for d in r:
                d.to_file(f)
        else:
            if self.filename and r.prefer_file_path:
                r.to_file(self.filename)
            else:
                r.to_file(self._f())


target = FileTarget



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import SimpleTarget

LOG = logging.getLogger(__name__)


class ZarrTarget(SimpleTarget):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._zarr_kwargs = kwargs
        self._ekd_kwargs = kwargs.pop("earthkit_to_xarray_kwargs", {})
        self._xr_kwargs = kwargs.pop("xarray_to_zarr_kwargs", {})
        self._encoder = "zarr"

    def close(self):
        """Close the target and flush the fdb.

        The target will not be able to write anymore.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        pass

    def flush(self):
        """Flush the fdb.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        pass

    def _write(self, data, **kwargs):
        r = self._encode(data, earthkit_to_xarray_kwargs=self._ekd_kwargs)
        ds = r.to_xarray()
        ds.to_zarr(**self._xr_kwargs)


target = ZarrTarget



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from .file import FileSource

LOG = logging.getLogger(__name__)


REMOTE_SAMPLE_DATA_URL = "https://get.ecmwf.int/repository/earthkit/samples/"


class Sample(FileSource):
    def __init__(self, path, *args, **kwargs):
        super().__init__()

        self._paths = path
        if isinstance(self._paths, str):
            self._paths = [path]

        self._kwargs = dict(**kwargs)

    def mutate(self):
        from earthkit.data.sources.url import Url

        urls = []
        for f_name in self._paths:
            urls.append(REMOTE_SAMPLE_DATA_URL + f_name)

        return Url(urls, **self._kwargs)


source = Sample



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.readers import Reader
from earthkit.data.readers.grib.index.file import GribFieldListInOneFile

LOG = logging.getLogger(__name__)


class GRIBReader(GribFieldListInOneFile, Reader):
    appendable = True  # GRIB messages can be added to the same file

    def __init__(self, source, path, parts=None, positions=None):
        _kwargs = {}
        for k in [
            # "array_backend",
            "grib_field_policy",
            "grib_handle_policy",
            "grib_handle_cache_size",
            "use_grib_metadata_cache",
        ]:
            _kwargs[k] = source._kwargs.get(k, None)

        for k in source._kwargs:
            if "-" in k:
                raise KeyError(f"Invalid option {k} in GRIBReader. Option names must not contain '-'.")

        Reader.__init__(self, source, path)
        GribFieldListInOneFile.__init__(self, path, parts=parts, positions=positions, **_kwargs)
        self._source_kwargs = source._kwargs

    def __repr__(self):
        return "GRIBReader(%s)" % (self.path,)

    def mutate_source(self):
        # A GRIBReader is a source itself
        return self

    def is_streamable_file(self):
        return True

    def __getstate__(self):
        from earthkit.data.core.config import CONFIG

        policy = CONFIG.get("grib-file-serialisation-policy")
        r = {"serialisation_policy": policy, "kwargs": self._source_kwargs}

        if policy == "path":
            r["path"] = self.path
            r["positions"] = self._positions
        else:
            r["messages"] = [f.message() for f in self]

        return r

    def __setstate__(self, state):
        policy = state["serialisation_policy"]
        if policy == "path":
            from earthkit.data import from_source

            path = state["path"]
            ds = from_source("file", path, **state["kwargs"])
            self.__init__(ds.source, path, positions=state["positions"])
        elif policy == "memory":
            from earthkit.data import from_source
            from earthkit.data.core.caching import cache_file

            def _create(path, args):
                with open(path, "wb") as f:
                    for message in state["messages"]:
                        f.write(message)

            path = cache_file(
                "GRIBReader",
                _create,
                [],
            )
            ds = from_source("file", path)
            self.__init__(ds.source, path)
        else:
            raise ValueError(f"Unknown serialisation policy {policy}")



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file


def test_grib_cube_core():
    ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    c = ds.cube("param", "level")

    assert c.user_shape == (3, 6)
    assert c.field_shape == (7, 12)
    assert c.extended_user_shape == (3, 6, 7, 12)
    assert c.count() == 18
    assert c.user_coords == {
        "param": ("t", "u", "v"),
        "level": (300, 400, 500, 700, 850, 1000),
    }

    # this slice is a field
    r = c[0, 0]
    assert r.shape == (7, 12)
    assert r.metadata(["param", "level"]) == ["t", 300]
    assert r.to_numpy().shape == (7, 12)
    assert np.isclose(r.to_numpy()[0, 0], 226.6531524658203)

    # this slice is a cube
    r = c[0, 0:2]
    assert r.user_shape == (1, 2)
    assert r.field_shape == (7, 12)
    assert r.extended_user_shape == (1, 2, 7, 12)
    assert r.user_coords == {"param": ("t",), "level": (300, 400)}
    assert r.count() == 2
    assert r.to_numpy().shape == (1, 2, 7, 12)
    assert np.isclose(r.to_numpy()[0, 0, 0, 0], 226.6531524658203)

    ref_meta = (["t", 300], ["t", 400])

    for i in range(len(ref_meta)):
        assert r[0, i].metadata(["param", "level"]) == ref_meta[i], f"{i=} ref_meta={ref_meta[i]}"

    # this slice is a cube
    r = c[1:3, 0:2]
    assert r.user_shape == (2, 2)
    assert r.field_shape == (7, 12)
    assert r.extended_user_shape == (2, 2, 7, 12)
    assert r.user_coords == {"param": ("u", "v"), "level": (300, 400)}
    assert r.count() == 4
    assert r.to_numpy().shape == (2, 2, 7, 12)
    assert np.isclose(r.to_numpy()[0, 0, 0, 0], 10.455490112304688)

    ref_meta = (["u", 300], ["u", 400], ["v", 300], ["v", 400])
    cnt = 0
    for par in range(2):
        for level in range(2):
            assert (
                r[par, level].metadata(["param", "level"]) == ref_meta[cnt]
            ), f"{cnt=} ref_meta={ref_meta[cnt]}"
            cnt += 1

    # this slice is a cube
    r = c[1, ...]
    assert r.user_shape == (1, 6)
    assert r.field_shape == (7, 12)
    assert r.extended_user_shape == (1, 6, 7, 12)
    assert r.user_coords == {"param": ("u",), "level": (300, 400, 500, 700, 850, 1000)}
    assert r.count() == 6
    assert r.to_numpy().shape == (1, 6, 7, 12)
    assert np.isclose(r.to_numpy()[0, 0, 0, 0], 10.455490112304688)

    ref_meta = (["u", 300], ["u", 400], ["u", 500], ["u", 700], ["u", 850], ["u", 1000])

    cnt = 0
    for par in range(1):
        for level in range(6):
            assert (
                r[par, level].metadata(["param", "level"]) == ref_meta[cnt]
            ), f"{cnt=} ref_meta={ref_meta[cnt]}"
            cnt += 1


def test_grib_cubelet():
    ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    c = ds.cube("param", "level")

    reading_chunks = None
    assert c.count(reading_chunks) == 18

    ic = []
    for i in range(3):
        for j in range(6):
            ic.append((i, j))

    ref = [
        226.6531524658203,
        244.00323486328125,
        255.8430633544922,
        271.26531982421875,
        272.53916931152344,
        272.5641784667969,
        10.455490112304688,
        6.213775634765625,
        4.7400054931640625,
        4.1455230712890625,
        -4.89837646484375,
        -6.2868804931640625,
        -3.07965087890625,
        -0.050384521484375,
        -0.9900970458984375,
        3.6385498046875,
        8.660964965820312,
        7.8334808349609375,
    ]

    for i, cb in enumerate(c.iterate_cubelets(reading_chunks)):
        assert cb.extended_icoords == ic[i]

    for i, cb in enumerate(c.iterate_cubelets(reading_chunks)):
        assert cb.to_numpy().shape == (7, 12)
        assert np.isclose(cb.to_numpy()[0, 0], ref[i])


def test_grib_cube_non_hypercube():
    ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    ds += from_source("file", earthkit_test_data_file("ml_data.grib"))[:2]
    assert len(ds) == 18 + 2

    with pytest.raises(ValueError):
        ds.cube("param", "level")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from abc import ABCMeta
from abc import abstractmethod
from importlib import import_module

LOG = logging.getLogger(__name__)


"""Assign encoders to file suffixes"""
_SUFFIXES = {
    (".grib", ".grb", ".grib1", ".grib2", ".grb1", ".grb2"): "grib",
    (".nc", ".nc3", ".nc4", ".netcdf"): "netcdf",
    (".tiff", ".tif"): "geotiff",
}


def _suffix_to_encoder(suffix):
    for suffixes, encoder in _SUFFIXES.items():
        if isinstance(suffixes, str):
            if suffix == suffixes:
                return encoder
        elif suffix in suffixes:
            return encoder


class EncodedData(metaclass=ABCMeta):
    """Base class for representing encoded data.

    It is meant to be used by a :class:`Target` to write/add data to a given target.
    It is the return value from the :meth:`Encoder.encode` method.
    """

    prefer_file_path = False

    @abstractmethod
    def to_bytes(self):
        """Return the data as a bytesarray"""
        pass

    @abstractmethod
    def to_file(self, f):
        """Write the data to a file.

        Parameters:
        -----------
        f: file-like object
            File-like object to write to
        """
        pass

    @abstractmethod
    def metadata(self, key=None):
        pass


class Encoder(metaclass=ABCMeta):
    """Base class for encoders.

    An encoder is used to encode data to a specific format that can be used by a :class:`Target`.

    Parameters:
    -----------
    template: obj, None
        The template to use to encode the data. Can be overridden in the :obj:`encode` method.
    metadata: dict, None
        Metadata to use when encoding the data. Can be overridden in the :obj:`encode` method.
    **kwargs:
        Additional keyword arguments
    """

    def __init__(self, template=None, metadata=None, **kwargs):
        self.template = template
        self.metadata = metadata or {}
        self.metadata.update(kwargs)

    @abstractmethod
    def encode(
        self,
        data=None,
        values=None,
        check_nans=False,
        metadata={},
        template=None,
        missing_value=9999,
        **kwargs,
    ) -> EncodedData:
        """Encode the data.

        Parameters:
        -----------
        data: obj, None
            The data to encode. Should be used via double dispatch. Must have an ``_encode()``
            method,  which will call the appropriate ``_encode_*`` method on the :class:`Encoder`.
        values: obj, None
            The values to encode.
        check_nans: bool
            If True, check for NaN values in the data and replace them with the ``missing_value``.
        metadata: dict
            Metadata to use when encoding the data. When None,
            the metadata the :class:`Encoder` was created with will be used if available.
        template: obj, None
            The template to use to encode the data. When None,
            the template the :class:`Encoder` was created with will be used if available.
        missing_value: number
            The value to use for missing values.
        **kwargs: dict
            Additional keyword arguments.

        Returns:
        --------
        EncodedData
            The encoded data.
        """
        pass

    @abstractmethod
    def _encode(self, data, **kwargs) -> EncodedData:
        """Subclass implementation of the encoding logic.

        Parameters:
        -----------
        field: :obj:`Base`
            The data to encode

        Double dispatch method that called from a ``data`` to encode itself."""
        pass

    @abstractmethod
    def _encode_field(self, field, **kwargs) -> EncodedData:
        """Subclass implementation of the encoding logic for a Field.

        Parameters:
        -----------
        field: :obj:`Field`
            The Field to encode

        Double dispatch method that called from ``field`` to encode itself."""
        pass

    @abstractmethod
    def _encode_fieldlist(self, fieldlist, **kwargs) -> EncodedData:
        """Subclass implementation of the encoding logic for a FieldList.

        Parameters:
        -----------
        fieldlist: :obj:`FieldList`
            The FieldList to encode

        Double dispatch method that called from ``fieldlist`` to encode itself."""
        pass

    @abstractmethod
    def _encode_xarray(self, data, **kwargs) -> EncodedData:
        """Subclass implementation of the encoding logic for Xarray data.

        Parameters:
        -----------
        data: Xarray DataArray or Dataset
            The data to encode

        Double dispatch method that called from ``data`` to encode itself."""
        pass


class EncoderLoader:
    kind = "encoder"

    def load_module(self, module):
        return import_module(module, package=__name__).encoder

    def load_entry(self, entry):
        entry = entry.load()
        if callable(entry):
            return entry
        return entry.encoder

    def load_remote(self, name):
        return None


class EncoderMaker:
    ENCODERS = {}

    def __call__(self, name_or_encoder, *args, **kwargs):
        if isinstance(name_or_encoder, Encoder):
            return name_or_encoder

        name = name_or_encoder

        loader = EncoderLoader()

        if name in self.ENCODERS:
            klass = self.ENCODERS[name]
        else:
            from earthkit.data.core.plugins import find_plugin

            klass = find_plugin(os.path.dirname(__file__), name, loader)
            self.ENCODERS[name] = klass

        encoder = klass(*args, **kwargs)

        if getattr(encoder, "name", None) is None:
            encoder.name = name

        return encoder

    def __getattr__(self, name):
        return self(name.replace("_", "-"))


create_encoder = EncoderMaker()


def make_encoder(data, encoder=None, suffix=None, metadata=None, **kwargs):
    if isinstance(encoder, Encoder):
        encoder.metadata.update(metadata or {})
        return encoder

    if encoder is None:
        if suffix is not None:
            encoder = _suffix_to_encoder(suffix)
        if encoder is None:
            if hasattr(data, "default_encoder"):
                # print("data.default_encoder", data.default_encoder())
                encoder = data.default_encoder()

    if isinstance(encoder, str):
        encoder = create_encoder(encoder, metadata=metadata, **kwargs)
        assert encoder is not None
        return encoder

    if encoder is not None:
        raise ValueError(f"Unsupported encoder={encoder}. Must be a str or Encoder")

    assert encoder is None

    raise ValueError("No data or encoder")



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data.utils.humanize import interval_to_human
from earthkit.data.utils.interval import Interval


@pytest.mark.parametrize("_kwargs", [{}, {"closed": "both"}])
def test_interval(_kwargs):
    v = Interval(1, 2, **_kwargs)
    assert 0 not in v
    assert 1 in v
    assert 1.5 in v
    assert 2 in v
    assert 3 not in v


def test_interval_closed_left():
    v = Interval(1, 2, closed="left")
    assert 0 not in v
    assert 1 in v
    assert 1.5 in v
    assert 2 not in v
    assert 3 not in v


def test_interval_closed_right():
    v = Interval(1, 2, closed="right")
    assert 0 not in v
    assert 1 not in v
    assert 1.5 in v
    assert 2 in v
    assert 3 not in v


def test_interval_closed_none():
    v = Interval(1, 2, closed="none")
    assert 0 not in v
    assert 1 not in v
    assert 1.5 in v
    assert 2 not in v
    assert 3 not in v


def test_interval_bounded_left():
    v = Interval(1, None)
    assert 0 not in v
    assert 1 in v
    assert 2 in v


def test_interval_bounded_right():
    v = Interval(None, 2)
    assert 1 in v
    assert 2 in v
    assert 3 not in v


def test_interval_unbounded():
    v = Interval(None, None)
    assert 1 in v
    assert 2 in v
    assert "a" in v


def test_interval_invalid():
    with pytest.raises(ValueError):
        Interval(2, 1)

    with pytest.raises(ValueError):
        Interval(1, 1)

    with pytest.raises(ValueError):
        Interval(2, 1, closed="all")

    with pytest.raises(TypeError):
        Interval("b", 2)


def test_interval_humanised():
    v = Interval(1, 2)
    assert interval_to_human(v) == "1 <= x <= 2"

    v = Interval(1, 2, closed="both")
    assert interval_to_human(v) == "1 <= x <= 2"

    v = Interval(1, 2, closed="left")
    assert interval_to_human(v) == "1 <= x < 2"

    v = Interval(1, 2, closed="right")
    assert interval_to_human(v) == "1 < x <= 2"

    v = Interval(1, 2, closed="none")
    assert interval_to_human(v) == "1 < x < 2"

    v = Interval(None, 2, closed="both")
    assert interval_to_human(v) == "x <= 2"

    v = Interval(None, 2, closed="left")
    assert interval_to_human(v) == "x < 2"

    v = Interval(None, 2, closed="right")
    assert interval_to_human(v) == "x <= 2"

    v = Interval(None, 2, closed="none")
    assert interval_to_human(v) == "x < 2"

    v = Interval(1, None, closed="both")
    assert interval_to_human(v) == "x >= 1"

    v = Interval(1, None, closed="left")
    assert interval_to_human(v) == "x >= 1"

    v = Interval(1, None, closed="right")
    assert interval_to_human(v) == "x > 1"

    v = Interval(1, None, closed="none")
    assert interval_to_human(v) == "x > 1"

    v = Interval(None, None)
    assert interval_to_human(v) == ""


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from functools import cached_property
from itertools import product

import deprecation

from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.index import MaskIndex
from earthkit.data.core.index import MultiIndex

from .coords import LevelCoordinate
from .coords import OtherCoordinate
from .coords import TimeCoordinate
from .dataset import GEOGRAPHIC_COORDS
from .dataset import DataSet
from .field import NetCDFField
from .field import XArrayField

LOG = logging.getLogger(__name__)


def get_fields_from_ds(
    ds,
    field_type=None,
    check_only=False,
):  # noqa C901

    fields = []

    skip = set()

    def _skip_attr(v, attr_name):
        attr_val = getattr(v, attr_name, "")
        if isinstance(attr_val, str):
            skip.update(attr_val.split(" "))

    for name in ds.data_vars:
        v = ds[name]
        _skip_attr(v, "coordinates")
        _skip_attr(v, "bounds")
        _skip_attr(v, "grid_mapping")

    for name in ds.data_vars:
        # Select only geographical variables
        has_lat = False
        has_lon = False

        if name in skip:
            continue

        v = ds[name]

        coordinates = []

        # self.log.info('Scanning file: %s var=%s coords=%s', self.path, name, v.coords)

        info = [value for value in v.coords if value not in v.dims]
        non_dim_coords = {}
        for coord in v.coords:
            if coord not in v.dims:
                non_dim_coords[coord] = ds[coord].values
                continue

            c = ds[coord]

            # self.log.info("COORD %s %s %s %s", coord, type(coord), hasattr(c, 'calendar'), c)

            standard_name = getattr(c, "standard_name", "")
            axis = getattr(c, "axis", "")
            long_name = getattr(c, "long_name", "")
            coord_name = getattr(c, "name", "")

            # LOG.debug(f"{standard_name=} {long_name=} {axis=} {coord_name}")
            use = False

            if (
                standard_name.lower() in GEOGRAPHIC_COORDS["x"]
                or (long_name == "longitude")
                or (axis == "X")
                or coord_name.lower() in GEOGRAPHIC_COORDS["x"]
            ):
                has_lon = True
                use = True

            if (
                standard_name.lower() in GEOGRAPHIC_COORDS["y"]
                or (long_name == "latitude")
                or (axis == "Y")
                or coord_name.lower() in GEOGRAPHIC_COORDS["y"]
            ):
                has_lat = True
                use = True

            # Of course, not every one sets the standard_name
            if (
                standard_name in ["time", "forecast_reference_time"]
                or long_name in ["time"]
                or coord_name.lower() in ["time"]
                or axis == "T"
            ):
                # we might not be able to convert time to datetime
                try:
                    coordinates.append(TimeCoordinate(c, coord in info))
                    use = True
                except ValueError:
                    break

            # TODO: Support other level types
            if (
                standard_name
                in [
                    "air_pressure",
                    "model_level_number",
                    "altitude",
                ]
                or long_name in ["pressure_level"]
                or coord_name in ["level"]
            ):  # or axis == 'Z':
                coordinates.append(LevelCoordinate(c, coord in info))
                use = True

            if axis in ("X", "Y"):
                use = True

            if not use:
                coordinates.append(OtherCoordinate(c, coord in info))

        if not (has_lat and has_lon):
            # self.log.info("NetCDFReader: skip %s (Not a 2 field)", name)
            continue

        for values in product(*[c.values for c in coordinates]):
            slices = []
            for value, coordinate in zip(values, coordinates):
                slices.append(coordinate.make_slice(value))

            if check_only:
                return True

            fields.append(field_type(ds, name, slices, non_dim_coords))

    # if not fields:
    #     raise Exception("NetCDFReader no 2D fields found in %s" % (self.path,))

    if check_only:
        return False

    return fields


class XArrayFieldListCore(FieldList):
    FIELD_TYPE = None

    def __init__(self, *args, **kwargs):
        self._fields = None
        super().__init__(*kwargs)

    # @cached_property
    @property
    def fields(self):
        if self._fields is None:
            self._fields = self._get_fields(DataSet(self.xr_dataset))
        return self._fields

    def has_fields(self):
        if self._fields is None:
            return get_fields_from_ds(
                DataSet(self.xr_dataset),
                field_type=self.FIELD_TYPE,
                check_only=True,
            )
        else:
            return len(self._fields) > 0

    def _get_fields(self, ds):
        return get_fields_from_ds(ds, field_type=self.FIELD_TYPE)

    def to_pandas(self, **kwargs):
        return self.to_xarray(**kwargs).to_pandas()

    def to_xarray(self, **kwargs):
        return self.xr_dataset

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))
        if not options:
            options = dict(**kwargs)

        return xr.open_mfdataset(
            paths,
            **options,
        )

    def to_netcdf(self, *args, **kwargs):
        """
        Save the data to a netCDF file.

        Parameters
        ----------
        See `xarray.DataArray.to_netcdf`.
        """
        return self.to_xarray().to_netcdf(*args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, XArrayFieldList) for _ in sources)
        return XArrayMultiFieldList(sources)

    @classmethod
    def new_mask_index(cls, *args, **kwargs):
        return XArrayMaskFieldList(*args, **kwargs)

    # def _write(self, target, **kwargs):
    #     assert target is not None
    #     encoder = kwargs.get("encoder", None)
    #     if encoder is None and kwargs.get("default_encoder", None) is None:
    #         kwargs["default_encoder"] = "netcdf"
    #     target._write(self, **kwargs)

    def _encode_grib(self, encoder, **kwargs):
        return encoder._from_xarray(self.to_xarray(), **kwargs)

    def default_encoder(self):
        return "netcdf"


class XArrayFieldList(XArrayFieldListCore):
    VERSION = 1

    def __init__(self, ds, **kwargs):
        self.FIELD_TYPE = XArrayField
        self._ds = ds
        super().__init__(**kwargs)

    @cached_property
    def xr_dataset(self):
        return self._ds

    def _getitem(self, n):
        if isinstance(n, int):
            return self.fields[n]

    def __len__(self):
        return len(self.fields)

    def __repr__(self):
        return "XArrayFieldList "  # TODO: some __repr__ of the data: {self.data}


class XArrayMaskFieldList(XArrayFieldListCore, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_mask(self, self)


class XArrayMultiFieldList(XArrayFieldListCore, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_multi(self, self)

    def to_xarray(self, **kwargs):
        import xarray as xr

        return xr.merge([x._ds for x in self._indexes], **kwargs)


class NetCDFFieldList(XArrayFieldListCore):
    FIELD_TYPE = NetCDFField

    def __init__(self, path, *args, **kwargs):
        self.path = path
        super().__init__(None, *args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, NetCDFFieldList) for _ in sources)
        return NetCDFMultiFieldList(sources)

    @classmethod
    def new_mask_index(cls, *args, **kwargs):
        return NetCDFMaskFieldList(*args, **kwargs)

    def to_xarray(self, **kwargs):
        # if self.path.startswith("http"):
        #     return xr.open_dataset(self.path, **kwargs)
        return type(self).to_xarray_multi_from_paths([self.path], **kwargs)

    # def save(self, *args, **kwargs):
    #     return self.to_netcdf(*args, **kwargs)

    # def write(self, *args, **kwargs):
    #     return self.to_netcdf(*args, **kwargs)


class NetCDFFieldListFromFileOrURL(NetCDFFieldList):
    def __init__(self, path_or_url, **kwargs):
        assert isinstance(path_or_url, str), path_or_url
        super().__init__(path_or_url, **kwargs)
        self.path_or_url = path_or_url

    @cached_property
    def xr_dataset(self):
        import xarray as xr

        return xr.open_dataset(self.path_or_url)

    def _getitem(self, n):
        if isinstance(n, int):
            return self.fields[n]

    def __len__(self):
        return len(self.fields)


class NetCDFFieldListFromFile(NetCDFFieldListFromFileOrURL):
    def __init__(self, path):
        super().__init__(path)

    def __repr__(self):
        return "NetCDFFieldListFromFile(%s)" % (self.path_or_url,)

    def write(self, f, **kwargs):
        import shutil

        with open(self.path, "rb") as s:
            shutil.copyfileobj(s, f, 1024 * 1024)


class NetCDFFieldListFromURL(NetCDFFieldListFromFileOrURL):
    def __init__(self, url):
        super().__init__(url)

    def __repr__(self):
        return "NetCDFFieldListFromURL(%s)" % (self.path_or_url,)


class NetCDFMaskFieldList(NetCDFFieldList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_mask(self, self)
        self.path = "<mask>"

    # TODO: Implement this, but discussion required
    def to_xarray(self, *args, **kwargs):
        self._not_implemented()

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def write(self, *args, **kwargs):
        self._not_implemented()

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def save(self, *args, **kwargs):
        self._not_implemented()


class NetCDFMultiFieldList(NetCDFFieldList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_multi(self, self)
        # self.paths = [s.path for s in args[0]]
        self.path = "<multi>"

    def to_xarray(self, **kwargs):
        # try:

        if not kwargs:
            kwargs = dict(combine="by_coords")

        for x in self._indexes:
            if isinstance(x, NetCDFMaskFieldList):
                raise NotImplementedError(
                    "NetCDFMultiFieldList.to_xarray() does not supports NetCDFMaskFieldList"
                )

        return NetCDFFieldList.to_xarray_multi_from_paths([x.path for x in self._indexes], **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import logging

from earthkit.data.utils import ensure_dict
from earthkit.data.utils.kwargs import Kwargs
from earthkit.data.utils.serialise import deserialise_state
from earthkit.data.utils.serialise import serialise_state

LOG = logging.getLogger(__name__)


class ItemWrapperForCfGrib:
    def __init__(self, item, ignore_keys=[]):
        self.item = item
        self.ignore_keys = ignore_keys

    def __getitem__(self, n):
        if n in self.ignore_keys:
            return None
        if n == "values":
            return self.item.values
        return self.item.metadata(n)


class IndexWrapperForCfGrib:
    def __init__(self, index=None, ignore_keys=[]):
        self.index = index
        self.ignore_keys = ignore_keys

    def __getstate__(self):
        return dict(index=serialise_state(self.index), ignore_keys=self.ignore_keys)

    def __setstate__(self, state):
        self.index = deserialise_state(state["index"])
        self.ignore_keys = state["ignore_keys"]

    def __getitem__(self, n):
        return ItemWrapperForCfGrib(
            self.index[n],
            ignore_keys=self.ignore_keys,
        )

    def __len__(self):
        return len(self.index)


class XarrayMixIn:

    @staticmethod
    def _kwargs_for_xarray():
        import inspect

        import xarray as xr

        r = inspect.signature(xr.open_dataset)
        v = []
        for p in r.parameters.values():
            if p.kind == p.KEYWORD_ONLY and p.name != "engine":
                v.append(p.name)
        return v

    def xarray_open_dataset_kwargs(self):
        return dict(
            cache=True,  # Set to false to prevent loading the whole dataset
            chunks=None,  # Set to 'auto' for lazy loading
        )

    def to_xarray(self, engine="earthkit", xarray_open_dataset_kwargs=None, **kwargs):
        """
        Convert the FieldList into an Xarray Dataset.

        Parameters
        ----------
        engine: str, optional
            The Xarray engine to use for generating the dataset. Default value is ``"earthkit"``.
            If set to ``cfgrib``, the :xref:`cfgrib` engine is used. No other values are
            supported.
        split_dims: str, or iterable of str, None
            Dimension or list of dimensions to use for splitting the data into multiple hypercubes.
            Default is None. Only used when ``engine="earthkit"``. Please note that ``split_dims``
            is not a valid option when the Xarray is directly generated via
            :py:meth:`xarray.open_dataset`.
        xarray_open_dataset_kwargs: dict, optional
            Keyword arguments passed to :py:func:`xarray.open_dataset`. Either this or ``**kwargs`` can
            be used, but not both.
        **kwargs: dict, optional
            Any keyword arguments that can be passed to :py:func:`xarray.open_dataset`. Engine specific
            keywords are automatically grouped and passed as ``backend_kwargs``.  Either ``**kwargs``
            or ``xarray_open_dataset_kwargs`` can be used, but not both.

            When ``engine="earthkit"`` the following engine specific kwargs are supported:

            * profile: str, dict or None
                Provide custom default values for most of the kwargs. Currently, the "mars" and "grid"
                built-in profiles are available, otherwise an explicit dict can be used. None is equivalent
                to an empty dict. When a kwarg is specified it will update
                a default value if it is a dict otherwise it will overwrite it. See: :ref:`xr_profile` for
                more information.
            * variable_key: str, None
                Metadata key to specify the dataset variables. It cannot be
                defined as a dimension. Default is "param" (in earthkit-data this is the same as "shortName").
                Only enabled when ``mono_variable`` is False or None.
            * drop_variables: str, or iterable of str, None
                A variable or list of variables to drop from the dataset. Default is None. Only used when
                ``variable_key`` is enabled.
            * rename_variables: dict, None
                Mapping to rename variables. Default is None. Only used  when
                ``variable_key`` is enabled.
            * mono_variable: bool, str, None
                If True or str, the dataset will contain a single variable called "data" (or the value
                of the ``mono_variable`` kwarg when it is a str). If False, the dataset will contain
                one variable for each distinct value of ``variable_key`` metadata key. The default value
                (None) expands to False unless the ``profile`` overwrites it.
            * extra_dims:  str, or iterable of str, None
                Define additional dimensions on top of the predefined dimensions. Only enabled when no
                ``fixed_dims`` is specified. Default is None. It can be a single item or a list. Each
                item is either a metadata key, or a dict/tuple defining mapping between the dimension
                name and the metadata key. The whole option can be a dict. E.g.

                .. code-block:: python

                    # use key "expver" as a dimension
                    extra_dims = "expver"
                    # use keys "expver" and "steam" as a dimension
                    extra_dims = ["expver", "stream"]
                    # define dimensions "expver", mars_stream" and "mars_type" from
                    # metadata keys "expver", "stream" and "type"
                    extra_dims = [
                        "expver",
                        {"mars_stream": "stream"},
                        ("mars_type", "type"),
                    ]
                    extra_dims = [
                        {
                            "expver": "expver",
                            "mars_stream": "stream",
                            "mars_type": "type",
                        }
                    ]

            * drop_dims:  str, or iterable of str, None
                Single or multiple dimensions to be ignored. Default is None.
                Default is None.
            * ensure_dims: str, or iterable of str, None
                Dimension or dimensions that should be kept even when ``squeeze=True`` and their size
                is only 1. Default is None.
            * fixed_dims: str, or iterable of str, None
                Define all the dimensions to be generated. When used no other dimensions will be created.
                Might be incompatible with other settings. Default is None. It can be a single item or a list.
                Each item is either a metadata key, or a dict/tuple defining mapping between the dimension
                name and the metadata key. The whole option can be a dict. E.g.:

                .. code-block:: python

                    # use key "step" as a dimension
                    fixed_dims = "step"
                    # use keys "step" and "levelist" as a dimension
                    extra_dims = ["step", "levelist"]
                    # define dimensions "step", level" and "level_type" from
                    # metadata keys "step", "levelist" and "levtype"
                    extra_dims = [
                        "step",
                        {"level": "levelist"},
                        ("level_type", "levtype"),
                    ]
                    extra_dims = [
                        {"step": "step", "level": "levelist", "level_type": "levtype"}
                    ]

            * dim_roles: dict, None
                Specify the "roles" used to form the predefined dimensions. The predefined dimensions are
                automatically generated when no ``fixed_dims`` specified and comprise the following
                (in a fixed order):

                - ensemble forecast member dimension
                - temporal dimensions (controlled by ``time_dim_mode``)
                - vertical dimensions (controlled by ``level_dim_mode``)

                ``dim_roles`` is a mapping between the "roles" and the metadata keys representing the roles.
                The possible roles are as follows:

                - "number": metadata key interpreted as ensemble forecast members
                - "date": metadata key interpreted as date part of the "forecast_reference_time"
                - "time": metadata key interpreted as time part of the "forecast_reference_time"
                - "step": metadata key interpreted as forecast step
                - "forecast_reference_time": if not specified or None or empty the forecast reference
                  time is built using the "date" and "time" roles
                - "valid_time": if not specified or None or empty the valid time is built using the
                  "validityDate" and "validityTime" metadata keys
                - "level": metadata key interpreted as level
                - "level_type": metadata key interpreted as level type

                The default values are as follows:

                .. code-block:: python

                    {
                        "number": "number",
                        "date": "dataDate",
                        "time": "dataTime",
                        "step": "step",
                        "forecast_reference_time": None,
                        "valid_date": None,
                        "level": "level",
                        "level_type": "typeOfLevel",
                    }

                ``dims_roles`` behaves differently to the other kwargs in the sense that
                it does not override but update the default values. So e.g. to change only "number" in
                the defaults it is enough to specify: "dim_roles={"number": "perturbationNumber"}.
            * dim_name_from_role_name: bool, None
                If True, the dimension names are formed from the role names. Otherwise the
                dimension names are formed from the metadata keys specified in ``dim_roles``.
                Its default value (None) expands to True unless the ``profile`` overwrites it.
                Only used when no `fixed_dims`` are specified. *New in version 0.15.0*.
            * rename_dims: dict, None
                Mapping to rename dimensions. Default is None.
            * dims_as_attrs: str, or iterable of str, None
                Dimension or list of dimensions which should be turned to variable
                attributes if they have only one value for the given variable. Default is None.
            * time_dim_mode: str, None
                Define how predefined temporal dimensions are formed. The default is "forecast".
                The possible values are as follows:

                - "forecast": adds two dimensions:

                  - "forecast_reference_time": built from the "date" and "time" roles
                    (see ``dim_roles``) as np.datetime64 values
                  - "step": built from the "step" role. When ``decode_time=True`` the values are
                    np.timedelta64
                - "valid_time": adds a dimension called "valid_time" as described by the "valid_time"
                  role (see ``dim_roles``). Will contain np.datetime64 values.
                - "raw": the "date", "time" and "step" roles are turned into 3 separate dimensions
            * level_dim_mode: str, None
                Define how predefined vertical dimensions are formed. The default is "level".
                The possible values are:

                - "level": adds a single dimension according to the "level" role (see ``dim_roles``)
                - "level_per_type": adds a separate dimensions for each level type based on the
                  "level" and "level_type" roles.
                - "level_and_type": Use a single dimension for combined level and type of level.
            * squeeze: bool, None
                Remove dimensions which have only one valid value. Not applies to dimensions in
                ``ensure_dims``. Its default value (None) expands
                to True unless the ``profile`` overwrites it.
            * add_valid_time_coord: bool, None
                If True, add the `valid_time` coordinate containing np.datetime64 values to the
                dataset. Only makes effect when ``time_dim_mode`` is not "valid_time". Its default
                value (None) expands to False unless the ``profile`` overwrites it.
            * decode_times: bool, None
                If True, decode date and datetime coordinates into datetime64 values. If False, leave
                coordinates representing date-like GRIB keys (e.g. "date", "validityDate") encoded as
                native int values. The default value (None) expands to True unless the ``profile``
                overwrites it.
            * decode_timedelta: bool, None
                If True, decode coordinates representing time-like or duration-like GRIB keys
                (e.g. "time", "validityTime", "step") into timedelta64 values. If False, leave time-like
                coordinates encoded as native int values, while duration-like coordinates will be encoded
                as int with the units attached to the coordinate as the "units" attribute.
                If None (default), assume the same value of ``decode_times`` unless the ``profile``
                overwrites it.
            * add_geo_coords: bool, None
                Add geographic coordinates to the dataset when field values are represented by
                a single "values" dimension. Its default value (None) expands
                to True unless the ``profile`` overwrites it.
            * flatten_values: bool, None
                If True, flatten the values per field resulting in a single dimension called
                "values" representing a field. If False, the field shape is used to form
                the field dimensions. When the fields are defined on an unstructured grid (e.g.
                reduced Gaussian) or are spectral (e.g. spherical harmonics) this option is
                ignored and the field values are always represented by a single "values"
                dimension.  Its default value (None) expands
                to False unless the ``profile`` overwrites it.
            * attrs_mode: str, None
                Define how attributes are generated. Default is "fixed". The possible values are:

                - "fixed": Use the attributes defined in ``variable_attrs`` as variables
                  attributes and ``global_attrs`` as global attributes.
                - "unique": Use all the attributes defined in ``attrs``, ``variable_attrs``
                  and ``global_attrs``.  When an attribute has unique value for a dataset
                  it will be a global attribute, otherwise it will be a variable attribute.
                  However, this logic is only applied if a unique variable attribute can be
                  a global attribute according to the CF conventions Appendix A. (e.g. "units" cannot
                  be a global attribute). Additionally, keys in ``variable_attrs`` are always used as
                  variable attributes, while keys in ``global_attrs`` are always used as global attributes.
            * attrs: str, number, callable, dict or list of these, None
                Attribute or list of attributes. Only used when ``attrs_mode`` is ``unique``.
                Its default value (None) expands to [] unless the ``profile`` overwrites it.
                The following attributes are supported:

                - str: Name of the attribute used as a metadata key to generate the value of
                  the attribute. Can also be specified by prefixing with "key=" (e.g. "key=level").
                  When prefixed with "namespace=" it specifies a metadata namespace
                  (e.g. "namespace=parameter"), which will be added as a dict to the attribute.
                - callable: A callable that takes a Metadata object and returns a dict of attributes
                - dict: A dictionary of attributes with the keys as the attribute names.
                  If the value is a callable it takes the attribute name and a Metadata object and
                  returns the value of the attribute. A str value prefixed with "key=" or
                  "namespace=" is interpreted as explained above. Any other values are used as the
                  pre-defined value for the attribute.
            * variable_attrs: str, number, callable, dict or list of these, None
                Variable attribute or attributes. For the allowed values see ``attrs``. Its
                default value (None) expands to [] unless the ``profile`` overwrites it.
            * global_attrs: str, number, dict or list of these, None
                Global attribute or attributes. For the allowed values see ``attrs``. Its
                default value (None) expands to [] unless the ``profile`` overwrites it.
            * coord_attrs: dict, None
                To be documented. Default is None.
            * add_earthkit_attrs: bool, None
                If True, add earthkit specific attributes to the dataset. Its default value
                (None) expands to True unless the ``profile`` overwrites it.
            * rename_attrs: dict, None
                A dictionary of attribute to rename. Default is None.
            * fill_metadata: dict, None
                Define fill_metadata values to metadata keys. Default is None.
            * remapping: dict, None
                Define new metadata keys for indexing. Default is None.
            * lazy_load: bool, None
                If True, the resulting Dataset will load data lazily from the
                underlying data source. If False, a Dataset holding all the data in memory
                and decoupled from the backend source will be created.
                Using ``lazy_load=False`` with ``release_source=True`` can provide optimised
                memory usage in certain cases. The default value of ``lazy_load`` (None)
                expands to True unless the ``profile`` overwrites it.
            * release_source: bool, None
                Only used when ``lazy_load=False``. If True, memory held in the input fields are
                released as soon as their values are copied into the resulting Dataset. This is
                done per field to avoid memory spikes. The release operation is currently
                only supported for GRIB fields stored entirely in memory, e.g. when read from a
                :ref:`stream <streams>`. When a field does not support the release operation, this
                option is ignored. Having run :obj:`to_xarray` the input data becomes unusable,
                so use this option carefully. The default value of ``release_source`` (None) expands
                to False unless the ``profile`` overwrites it.
            * strict: bool, None
                If True, perform stricter checks on hypercube consistency. Its default value (None) expands
                to False unless the ``profile`` overwrites it.
            * dtype: str, numpy.dtype or None
                Typecode or data-type of the array data.
            * array_backend: str, array namespace, ArrayBackend, None
                The array backend/namespace to use for array operations. The default value (None) is
                expanded to "numpy".
            * direct_backend: bool, None
                If True, the backend is used directly bypassing :py:meth:`xarray.open_dataset`
                and ignoring all non-backend related kwargs. If False, the data is read via
                :py:meth:`xarray.open_dataset`. Its default value (None) expands
                to False unless the ``profile`` overwrites it.

            When ``engine="cfgrib"`` the following engine specific kwargs are supported:

            * ignore_keys: list, None
                It specifies the metadata keys that should be ignored when reading
                the GRIB messages in the backend. Please note that is not supported by :xref:`cfgrib`, but
                implemented in earthkit-data.
            * For the rest of the supported keyword arguments, please refer to the
              :xref:`cfgrib` documentation.


        Returns
        -------
        Xarray Dataset or tuple
            When ``split_dims`` is unset a Dataset is returned. When ``engine="earthkit"``
            and ``split_dims`` is set a tuple is returned. The first element of the tuple is the list
            of Datasets and the second element is the list of corresponding dictionaries with
            the spitting keys/values (one dictionary per Dataset).

        Notes
        -----
        The default values of ``xarray_open_dataset_kwargs`` or ``**kwargs`` passed
        to :py:func:`xarray.open_dataset` are as follows:

            - when ``engine="earthkit"``::

               {"cache": True, "chunks": None, "engine": "earthkit"}

            - when ``engine="cfgrib"``::

                {"backend_kwargs": {"errors": "raise", "ignore_keys": [], "squeeze": False,},
                "cache": True, "chunks": None, "engine": "cfgrib"}

        Please note that settings ``errors="raise"`` and ``engine`` are always enforced
        and cannot be changed.


        Examples
        --------
        >>> import earthkit.data
        >>> fs = earthkit.data.from_source("file", "test6.grib")
        >>> ds = fs.to_xarray(time_dim_mode="forecast")
        >>> # also possible to use the xarray_open_dataset_kwargs
        >>> ds = fs.to_xarray(
        ...     xarray_open_dataset_kwargs={
        ...         "backend_kwargs": {"time_dim_mode": "forecast"}
        ...     }
        ... )

        """
        engines = {"earthkit": self.to_xarray_earthkit, "cfgrib": self.to_xarray_cfgrib}
        if engine not in engines:
            raise ValueError(f"Unsupported engine: {engine}. Please use one of {list(engines.keys())}")

        user_xarray_open_dataset_kwargs = ensure_dict(xarray_open_dataset_kwargs)
        user_xarray_open_dataset_kwargs = copy.deepcopy(user_xarray_open_dataset_kwargs)
        if user_xarray_open_dataset_kwargs and kwargs:
            raise ValueError("Cannot specify extra keyword arguments when xarray_open_dataset_kwargs is set.")

        if not user_xarray_open_dataset_kwargs:
            user_xarray_open_dataset_kwargs = dict(**kwargs)

        backend_kwargs = user_xarray_open_dataset_kwargs.pop("backend_kwargs", {})
        xr_kwargs = self._kwargs_for_xarray()
        for key in list(user_xarray_open_dataset_kwargs.keys()):
            if key not in xr_kwargs:
                if key in backend_kwargs:
                    raise ValueError(f"Duplicate keyword argument={key}")
                backend_kwargs[key] = user_xarray_open_dataset_kwargs.pop(key)
        user_xarray_open_dataset_kwargs["backend_kwargs"] = backend_kwargs

        return engines[engine](user_xarray_open_dataset_kwargs)

    def to_xarray_earthkit(self, user_kwargs):
        xarray_open_dataset_kwargs = {}

        for key in ["backend_kwargs"]:
            xarray_open_dataset_kwargs[key] = Kwargs(
                user=user_kwargs.pop(key, {}),
                default={"errors": "raise"},
                forced={},
                logging_owner="xarray_open_dataset_kwargs",
                logging_main_key=key,
            )

        # print(f"{xarray_open_dataset_kwargs=}")

        default = dict()
        default.update(self.xarray_open_dataset_kwargs())

        xarray_open_dataset_kwargs.update(
            Kwargs(
                user=user_kwargs,
                default=default,
                forced={
                    # "errors": "raise",
                    "engine": "earthkit",
                },
                logging_owner="xarray_open_dataset_kwargs",
                warn_non_default=False,
            )
        )

        # print(f"{kwargs=}")
        # print(f"{xarray_open_dataset_kwargs=}")

        # separate backend_kwargs from other_kwargs
        backend_kwargs = xarray_open_dataset_kwargs.pop("backend_kwargs", None)
        other_kwargs = xarray_open_dataset_kwargs

        from earthkit.data.utils.xarray.builder import from_earthkit

        return from_earthkit(self, backend_kwargs=backend_kwargs, other_kwargs=other_kwargs)

    def to_xarray_cfgrib(self, user_kwargs):
        xarray_open_dataset_kwargs = {}

        # until ignore_keys is included into cfgrib,
        # it is implemented here directly
        ignore_keys = user_kwargs.get("backend_kwargs", {}).pop("ignore_keys", [])

        for key in ["backend_kwargs"]:
            xarray_open_dataset_kwargs[key] = Kwargs(
                user=user_kwargs.pop(key, {}),
                default={"errors": "raise", "squeeze": False},
                forced={},
                logging_owner="xarray_open_dataset_kwargs",
                logging_main_key=key,
            )

        # default = dict(squeeze=False)  # TODO:Document me
        default = dict()
        default.update(self.xarray_open_dataset_kwargs())

        xarray_open_dataset_kwargs.update(
            Kwargs(
                user=user_kwargs,
                default=default,
                forced={
                    # "errors": "raise",
                    "engine": "cfgrib",
                },
                logging_owner="xarray_open_dataset_kwargs",
                warn_non_default=False,
            )
        )

        open_object = IndexWrapperForCfGrib(self, ignore_keys=ignore_keys)

        import xarray as xr

        return xr.open_dataset(
            open_object,
            # decode_cf=False,
            **xarray_open_dataset_kwargs,
        )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_S3_AUTH


def _check_s3_credentials():
    import os

    key_id = os.environ.get("AWS_ACCESS_KEY_ID", None)
    key = os.environ.get("AWS_SECRET_ACCESS_KEY", None)
    if key_id and key:
        return True
    return False


NO_S3_CREDENTIALS = not _check_s3_credentials()


@pytest.mark.parametrize(
    "objects,expected_len",
    [
        ("test6.grib", 6),
        ({"object": "test6.grib"}, 6),
        ([{"object": "test6.grib"}], 6),
        ([{"object": "test6.grib"}, {"object": "tuv_pl.grib"}], 24),
    ],
)
def test_s3_ewc_request(objects, expected_len):
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data-public",
        "objects": objects,
    }

    ds = from_source("s3", r, stream=False, anon=True)
    assert len(ds) == expected_len


def test_s3_ewc_public_core():
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data-public",
        "objects": [{"object": "test6.grib"}],
    }

    ds = from_source("s3", r, stream=False, anon=True)

    assert len(ds) == 6

    ref = [
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]

    assert ds.metadata(("param", "level")) == ref


@pytest.mark.skipif(NO_S3_AUTH, reason="No S3 authenticator")
@pytest.mark.skipif(NO_S3_CREDENTIALS, reason="No S3 credentials available")
def test_s3_ewc_private_core():
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data",
        "objects": [{"object": "test6.grib"}],
    }

    ds = from_source("s3", r, stream=False, anon=False)

    assert len(ds) == 6

    ref = [
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]

    assert ds.metadata(("param", "level")) == ref


@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ((0, 150), [("t", 1000)]),
        ((240, 150), [("u", 1000)]),
        ((240, 480), [("u", 1000), ("v", 1000)]),
    ],
)
def test_s3_ewc_public_single_parts(parts, expected_meta):
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data-public",
        "objects": [{"object": "test6.grib", "parts": parts}],
    }

    ds = from_source("s3", r, stream=False, anon=True)

    assert len(ds) == len(expected_meta)
    assert ds.metadata(("param", "level")) == expected_meta


@pytest.mark.skipif(NO_S3_AUTH, reason="No S3 authenticator")
@pytest.mark.skipif(NO_S3_CREDENTIALS, reason="No S3 credentials available")
@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ((0, 150), [("t", 1000)]),
        ((240, 150), [("u", 1000)]),
        ((240, 480), [("u", 1000), ("v", 1000)]),
    ],
)
def test_s3_ewc_private_single_parts(parts, expected_meta):
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data",
        "objects": [{"object": "test6.grib", "parts": parts}],
    }

    ds = from_source("s3", r, stream=False, anon=False)

    assert len(ds) == len(expected_meta)
    assert ds.metadata(("param", "level")) == expected_meta


def test_s3_ewc_public_stream_core():
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data-public",
        "objects": [{"object": "test6.grib"}],
    }

    ds = from_source("s3", r, stream=True, anon=True)

    # no fieldlist methods are available
    with pytest.raises(NotImplementedError):
        len(ds)

    ref = [
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.skipif(NO_S3_AUTH, reason="No S3 authenticator")
@pytest.mark.skipif(NO_S3_CREDENTIALS, reason="No S3 credentials available")
def test_s3_ewc_private_stream():
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data",
        "objects": [{"object": "test6.grib"}],
    }

    ds = from_source("s3", r, stream=True, anon=False)

    # no fieldlist methods are available
    with pytest.raises(NotImplementedError):
        len(ds)

    ref = [
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ((0, 150), [("t", 1000)]),
        ((240, 150), [("u", 1000)]),
        ((240, 480), [("u", 1000), ("v", 1000)]),
    ],
)
def test_s3_ewc_public_stream_single_parts(parts, expected_meta):
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data-public",
        "objects": [{"object": "test6.grib", "parts": parts}],
    }

    ds = from_source("s3", r, stream=True, anon=True)

    # no fieldlist methods are available
    with pytest.raises(NotImplementedError):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.skipif(NO_S3_AUTH, reason="No S3 authenticator")
@pytest.mark.skipif(NO_S3_CREDENTIALS, reason="No S3 credentials available")
@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ((0, 150), [("t", 1000)]),
        ((240, 150), [("u", 1000)]),
        ((240, 480), [("u", 1000), ("v", 1000)]),
    ],
)
def test_s3_ewc_private_stream_single_parts(parts, expected_meta):
    r = {
        "endpoint": "object-store.os-api.cci1.ecmwf.int",
        "bucket": "earthkit-test-data",
        "objects": [{"object": "test6.grib", "parts": parts}],
    }

    ds = from_source("s3", r, stream=True, anon=False)

    # no fieldlist methods are available
    with pytest.raises(NotImplementedError):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def test_file_pattern_source_grib():
    ds = from_source("file-pattern", earthkit_examples_file("test{id}.grib"), {"id": [4, 6]})

    assert len(ds) == 10
    assert ds.metadata("param") == ["t", "z", "t", "z", "t", "u", "v", "t", "u", "v"]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file

all_params = [
    "latitude",
    "sin_latitude",
    "cos_latitude",
    "longitude",
    "sin_longitude",
    "cos_longitude",
    "local_time",
    "sin_local_time",
    "cos_local_time",
    "julian_day",
    "sin_julian_day",
    "cos_julian_day",
    "ecef_x",
    "ecef_y",
    "ecef_z",
    "toa_incident_solar_radiation",
    "cos_solar_zenith_angle",
]


def load_forcings_fs(params=None, first_step=6, last_step=72, input_data="grib"):
    sample = from_source("file", earthkit_examples_file("test.grib"))

    if params is None:
        params = [
            "sin_latitude",
            "cos_latitude",
            "longitude",
            "sin_longitude",
            "cos_longitude",
            "local_time",
            "sin_local_time",
            "cos_local_time",
        ]

    start = sample[0].datetime()["valid_time"]
    step_increment = 6
    dates = []
    for step in range(first_step, last_step + step_increment, step_increment):
        dates.append(start + datetime.timedelta(hours=step))

    if input_data == "grib":
        ds = from_source(
            "forcings",
            sample,
            date=dates,
            param=params,
        )
    elif input_data == "latlon":
        d = {}
        ll = sample[0].to_latlon()
        d["latitudes"] = ll["lat"]
        d["longitudes"] = ll["lon"]
        # d["date"] = sample[0].metadata("date")
        # d["param"] = sample[0].metadata("param")
        ds = from_source(
            "forcings",
            **d,
            date=dates,
            param=params,
        )
    else:
        raise ValueError(f"Unknown input_data: {input_data}")

    assert len(ds) == len(dates) * len(params)

    md = [[d.isoformat(), p] for d, p in itertools.product(dates, params)]
    return ds, md



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import Reader

LOG = logging.getLogger(__name__)


class ODBReader(Reader):
    def to_pandas(self, **kwargs):
        try:
            import codc as odc
        except Exception:
            try:
                import pyodc as odc
            except ImportError:
                raise ImportError("ODC handling requires 'pyodc' to be installed")
            LOG.debug("Using pure Python odc decoder.")

        odc_read_odb_kwargs = kwargs.get("odc_read_odb_kwargs", {})
        return odc.read_odb(self.path, single=True, **odc_read_odb_kwargs)


def _match_magic(magic, deeper_check):
    if magic is not None:
        type_id = b"\xff\xffODA"
        if not deeper_check:
            return len(magic) >= 5 and magic[:5] == type_id
        else:
            return type_id in magic
    return False


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    if _match_magic(magic, deeper_check):
        return ODBReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import os

from earthkit.data.readers.grib.index.sql import FieldListInFilesWithSqlIndex
from earthkit.data.readers.grib.parsing import GribIndexingPathParserIterator
from earthkit.data.sources.indexed import IndexedSource

LOG = logging.getLogger(__name__)


def make_absolute(filename, root_dir, default):
    if filename is None:
        filename = default

    if os.path.isabs(filename):
        return filename

    absolute = os.path.join(root_dir, filename)
    LOG.debug(f"Transforming {filename} into absolute path {absolute}")
    return absolute


class FileIndexedSource(IndexedSource):
    DEFAULT_JSON_FILE = "earthkit.index"
    DEFAULT_DB_FILE = "earthkit.db"
    INDEX_CLASS = FieldListInFilesWithSqlIndex

    def __init__(
        self,
        path,
        db_path=None,
        index_file=None,
        _index=None,
        **kwargs,
    ):
        """
        index_file: optional
            The input index location. Must be a JSON file.
            If None, parse the actual files in the directory pointed by "path".
        db_path: optional
            The actual used database file. Must be a SQL earthkit index file.
            If None, create one in the default location.

        If _index is not None, ignore all other arguments (for internal usage)
        """
        if _index is not None:  # for .sel()
            super().__init__(_index, **kwargs)
            return

        path = os.path.expanduser(path)

        self.path = path
        self.abspath = os.path.abspath(path)

        if db_path is None:
            db_path = make_absolute(
                db_path,
                self.abspath,
                default=self.DEFAULT_DB_FILE,
            )

        # Try to use db_path if it exists:
        if os.path.exists(db_path):
            LOG.info(f"Using index file {db_path}")
            index = self.INDEX_CLASS.from_existing_db(db_path=db_path)
            super().__init__(index, **kwargs)
            return

        index_file = make_absolute(
            index_file,
            self.abspath,
            default=self.DEFAULT_JSON_FILE,
        )
        assert db_path != index_file

        # Try to use index_file (json) if it exists:
        if os.path.exists(index_file):
            LOG.info(f"Using index file {index_file}")
            print(f"Using index file {index_file} (will happen only once).")
            index = self.INDEX_CLASS.from_file(path=index_file)
            super().__init__(index, **kwargs)
            return

        # Create the db_path file in cache (or used the cached one)
        LOG.info(f"Did not find index files in {db_path} or {index_file}")
        ignore = [self.DEFAULT_DB_FILE, self.DEFAULT_JSON_FILE, db_path, index_file]
        index = self.INDEX_CLASS.from_iterator(
            GribIndexingPathParserIterator(path, ignore=ignore, relative_paths=False),
            cache_metadata={"directory": self.path},
        )

        super().__init__(index, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import logging
import os
import re
from functools import cached_property
from pathlib import Path
from typing import Any as TypingAny
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from .dates import to_datetime
from .dates import to_timedelta

LOG = logging.getLogger(__name__)

RE1 = re.compile(r"{([^}]*)}")
RE2 = re.compile(r"\(([^}]*)\)")


class Any:
    """Represents a type that accepts any value."""

    def substitute(self, value: TypingAny, name: str) -> TypingAny:
        """Substitute the value without any validation.

        Parameters
        ----------
        value : Any
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        Any
            The substituted value.
        """
        return value


class Enum:
    """Represents a type that accepts a value from a predefined set of strings.

    Parameters
    ----------
    enum : str, optional
        Comma-separated string of allowed values, by default "".
    """

    def __init__(self, enum: str = "") -> None:
        self.enum: set[str] = set(enum.split(","))

    def substitute(self, value: str, name: str) -> str:
        """Substitute the value if it is in the predefined set.

        Parameters
        ----------
        value : str
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted value.

        Raises
        ------
        ValueError
            If the value is not in the predefined set.
        """
        if self.enum and value not in self.enum:
            raise ValueError(
                "Invalid value '{}' for parameter '{}', expected one of {}".format(value, name, self.enum)
            )
        return value


class Int:
    """Represents a type that accepts integer values.

    Parameters
    ----------
    format : str, optional
        Format string for integer substitution, by default "%d".
    """

    def __init__(self, format: str = "%d") -> None:
        self.format: str = format

    def substitute(self, value: int, name: str) -> str:
        """Substitute the value if it is an integer.

        Parameters
        ----------
        value : int
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted value.

        Raises
        ------
        ValueError
            If the value is not an integer.
        """
        if not isinstance(value, int):
            raise ValueError("Invalid value '{}' for parameter '{}', expected an integer".format(value, name))
        return self.format % value


class Float:
    """Represents a type that accepts float values.

    Parameters
    ----------
    format : str, optional
        Format string for float substitution, by default "%g".
    """

    def __init__(self, format: str = "%g") -> None:
        self.format: str = format

    def substitute(self, value: Union[int, float], name: str) -> str:
        """Substitute the value if it is a float or integer.

        Parameters
        ----------
        value : int or float
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted value.

        Raises
        ------
        ValueError
            If the value is not a float or integer.
        """
        if not isinstance(value, (int, float)):
            raise ValueError("Invalid value '{}' for parameter '{}', expected a float".format(value, name))

        return self.format % value


class Datetime:
    """Represents a type that accepts datetime values.

    Parameters
    ----------
    format : str
        Format string for datetime substitution.
    """

    def __init__(self, format: str) -> None:
        self.format: str = format

    def substitute(self, value: TypingAny, name: str) -> str:
        """Substitute the value as a formatted datetime string.

        Parameters
        ----------
        value : Any
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted datetime string.
        """
        return to_datetime(value).strftime(self.format)


class DatetimeDelta:
    """Represents a type that accepts datetime deltas.

    Parameters
    ----------
    params : str
        Parameters for delta and format, separated by a semicolon.

    Attributes
    ----------
    delta : str
        Timedelta string for datetime substitution. Accepted formats use the following suffixes:
        - "h" for hours (the default)
        - "m" for minutes
        - "s" for seconds

    format : str
        Format string for datetime substitution.
    """

    def __init__(self, params: str) -> None:
        params_list = params.split(";")
        if len(params_list) != 2:
            raise ValueError(
                "Invalid parameters '{}' for class DatetimeDelta, expected (delta;format)".format(params)
            )
        self.delta = params_list[0].strip()
        self.format = params_list[1].strip()

    def substitute(self, value: TypingAny, name: str) -> str:
        """Substitute the datetime value, add the delta and format the result.

        Parameters
        ----------
        value : Any
            The value to substitute. Must be convertible to a datetime object.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted datetime value with the delta added to it and formatted.

        Raises
        ------
        ValueError
            If the delta format is invalid.
        """
        sign = -1 if self.delta[0] == "-" else 1
        if re.fullmatch(r"[-\+]?\d+[hms]?", self.delta):
            delta = re.search(r"\d+[hms]?", self.delta).group(0)
        else:
            raise ValueError(
                "Invalid value '{}' for delta, expected time in hour (h), minute (m) or second (s)".format(
                    self.delta
                )
            )

        valid_date = to_datetime(value) + sign * to_timedelta(delta)
        return valid_date.strftime(self.format)


class Str:
    """Represents a type that accepts string values.

    Parameters
    ----------
    format : str, optional
        Format string for string substitution, by default "%s".
    """

    def __init__(self, format: str = "%s") -> None:
        self.format: str = format

    def substitute(self, value: str, name: str) -> str:
        """Substitute the value if it is a string.

        Parameters
        ----------
        value : str
            The value to substitute.
        name : str
            The name of the parameter.

        Returns
        -------
        str
            The substituted value.

        Raises
        ------
        ValueError
            If the value is not a string.
        """
        if not isinstance(value, str):
            raise ValueError("Invalid value '{}' for parameter '{}', expected a string".format(value, name))
        return self.format % value


TYPES = {
    "": Any,
    "int": Int,
    "float": Float,
    "date": Datetime,
    "strftime": Datetime,
    "strftimedelta": DatetimeDelta,
    "enum": Enum,
}


class Constant:
    """Represents a constant value in a pattern.

    Parameters
    ----------
    value : Any
        The constant value.
    """

    name: Optional[str] = None

    def __init__(self, value: TypingAny) -> None:
        self.value: TypingAny = value

    def substitute(self, params: Dict[str, TypingAny], **kwargs: TypingAny) -> TypingAny:
        """Substitute the constant value.

        Parameters
        ----------
        params : dict
            Parameters to substitute (not used for constants).

        Returns
        -------
        Any
            The constant value.
        """
        return self.value

    def substitute_many(self, params: Dict[str, TypingAny], **kwargs: TypingAny) -> TypingAny:
        return [self.value]

    def pattern(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value})"


class Variable:
    """Represents a variable in a pattern.

    Parameters
    ----------
    value : str
        The variable definition string.
    """

    def __init__(self, value: str) -> None:
        bits = value.split(":")
        self.name = bits[0]
        kind = RE2.split(":".join(bits[1:]))
        if len(kind) == 1:
            self.kind = TYPES[kind[0]]()
        else:
            self.kind = TYPES[kind[0]](kind[1])

        self.value = value

    def substitute(self, params: Dict[str, TypingAny]) -> TypingAny:
        """Substitute value for a parameter.

        Parameters
        ----------
        params : dict
            The value belonging to ``self.name`` in ``params`` are
            substituted into the Variable. The value to substitute must be a
            single value.

        Returns
        -------
            Substituted value.

        Raises
        ------
        ValueError
            If ``self.name`` is not in ``params``.

        Example
        -------
        >>> v = Variable("my_date:date(%Y-%m-%d)")
        >>> v.substitute({"my_date": "2000-01-01"})
        '2000-01-01'

        >>> v = Variable("my_date:date(%Y-%m-%d)")
        >>> v.substitute({"level": "500"})
        ValueError: Missing parameter 'my_date'
        """
        if self.name in params:
            return self.kind.substitute(params[self.name], self.name)
        else:
            raise ValueError("Missing parameter '{}'".format(self.name))

    def substitute_many(self, params: Dict[str, TypingAny]) -> Optional[List[TypingAny]]:
        """Substitute all values for a parameter.

        Parameters
        ----------
        params : dict
            The values belonging to ``self.name`` in ``params`` are
            substituted into the Variable.

        Returns
        -------
        list
            List of substituted values.
        Raises
        ------
        ValueError
            If ``self.name`` is not in ``params``.


        Example
        -------
        >>> v = Variable("my_date:date(%Y-%m-%d)")
        >>> v.substitute_many({"my_date": ["2000-01-01", "2000-01-02"]})
        ['2000-01-01', '2000-01-02']

        >>> v = Variable("my_date:date(%Y-%m-%d)")
        >>> v.substitute_many({"my_date": "2000-01-01"})
        ['2000-01-01']

        >>> v = Variable("my_date:date(%Y-%m-%d)")
        >>> v.substitute_many({"level": "500"})
        ValueError: Missing parameter 'my_date'
        """
        if self.name in params:
            v = params[self.name]
            if not isinstance(v, list):
                v = [v]
            return [self.kind.substitute(x, self.name) for x in v]
        else:
            raise ValueError("Missing parameter '{}'".format(self.name))

    def pattern(self) -> str:
        return "{" + self.value + "}"

    def __repr__(self) -> str:
        return f"Variable({self.name},{self.value},{self.kind})"


FUNCTIONS = dict(lower=lambda s: s.lower())


class Function:
    """Represents a function applied to a variable in a pattern.

    Parameters
    ----------
    value : str
        The function definition string.
    """

    def __init__(self, value: str) -> None:
        functions = value.split("|")
        self.name = functions[0]
        self.variable = Variable(functions[0])
        self.functions = functions[1:]
        self._pattern = value

    def substitute(self, params: Dict[str, TypingAny], **kwargs: TypingAny) -> TypingAny:
        """Substitute the variable and apply functions.

        Parameters
        ----------
        params : dict
            Parameters to substitute.

        Returns
        -------
        Any
            The substituted and transformed value.
        """
        value = self.variable.substitute(params, **kwargs)
        for f in self.functions:
            value = FUNCTIONS[f](value)
        return value

    def substitute_many(self, params: Dict[str, TypingAny], **kwargs: TypingAny) -> TypingAny:
        value = self.variable.substitute_many(params, **kwargs)
        res = []
        for f in self.functions:
            for v in value:
                res.append(FUNCTIONS[f](v))
        return res

    def pattern(self) -> str:
        return self._pattern


class Pattern:
    """Represents a pattern with variables and constants.

    Parameters
    ----------
    pattern : str
        The pattern string.
    """

    def __init__(self, pattern: str) -> None:
        self.pattern = []
        self.variables = []
        for i, p in enumerate(RE1.split(pattern)):
            if i % 2 == 0:
                if p != "":
                    self.pattern.append(Constant(p))
            else:
                if "|" in p:
                    v = Function(p)
                else:
                    v = Variable(p)
                self.variables.append(v)
                self.pattern.append(v)

    @property
    def names(self) -> List[str]:
        return sorted({v.name for v in self.variables})

    def is_constant(self) -> bool:
        return not self.variables and len(self.pattern) == 1 and isinstance(self.pattern[0], Constant)

    def substitute(
        self,
        *args: Tuple[Dict[str, TypingAny]],
        allow_extra: bool = False,
        **kwargs: TypingAny,
    ) -> Union[str, List[str]]:
        """Substitute values into the pattern.

        Parameters
        ----------
        args : tuple of dict
            Positional dictionaries of parameters to substitute.
        allow_extra : bool, optional
            Whether to allow using input with parameters not part of the Pattern.
        kwargs : dict
            Additional keyword arguments for substitution.

        Returns
        -------
        str or list
            The substituted pattern as a string or list of strings.

        Raises
        ------
        ValueError
            If there are unused parameters and `allow_extra` is False.
        """
        params = {}
        for a in args:
            params.update(a)
        params.update(kwargs)

        for k, v in params.items():
            if isinstance(v, list):
                return self._substitute_many(params, allow_extra=allow_extra)

        return self._substitute_one(params, allow_extra=allow_extra)

    def _substitute_one(
        self,
        params: Dict[str, TypingAny],
        allow_extra: bool = False,
    ) -> str:
        used = set(params.keys())
        result = []
        for p in self.pattern:
            used.discard(p.name)
            result.append(p.substitute(params))
        if used and not allow_extra:
            raise ValueError("Unused parameter(s): {}".format(used))

        return "".join(str(x) for x in result)

    def _substitute_many(
        self,
        params: Dict[str, TypingAny],
        allow_extra: bool = False,
    ) -> List[str]:
        for k, v in list(params.items()):
            if not isinstance(v, list):
                params[k] = [v]

        seen = set()
        result = []
        for n in (dict(zip(params.keys(), x)) for x in itertools.product(*params.values())):
            m = self.substitute(n, allow_extra=allow_extra)
            if m not in seen:
                seen.add(m)
                result.append(m)

        return result

    def _subpattern(self, params: Dict[str, TypingAny]) -> str:
        """Substitute the pattern with the given parameters.

        Parameters
        ----------
        params : dict
            The parameters to substitute. Each parameter must be a single value.

        Returns
        -------
        str
            The substituted pattern.
        """
        result = []
        for p in self.pattern:
            try:
                result.append(p.substitute(params))
            except ValueError:
                result.append(p.pattern())

        return "".join(str(x) for x in result)

    def match(self, value: str) -> Optional[re.Match]:
        """Match pattern regex against value.

        Parameters
        ----------
        value : str
            Value to match

        Returns
        -------
        re.Match
            re.Match object if the value matches the pattern. None otherwise.

        Example
        -------
        >>> p = Pattern("t_{my_date:date(%Y-%m-%d)}.grib")
        >>> p.match("t_2000-01-01.grib")
        <re.Match object; span=(0, 17), match='t_2000-01-01.grib'>

        >>> p = Pattern("t_{my_date:date(%Y-%m-%d)}.grib")
        >>> p.match("2000-01-01.grib")
        None

        >>> p = Pattern("{shortName}_{my_date:date(%Y-%m-%d)}.grib")
        >>> p.match("t_2000-01-01.grib")
        <re.Match object; span=(0, 17), match='t_2000-01-01.grib'>

        >>> p = Pattern("data/t/level")
        >>> p.match("data/t/level")
        <re.Match object; span=(0, 12), match='data/t/level'>
        >>> p.match("data/t/level/500")
        None
        """
        rx = self.regex
        return rx.match(value)

    def __repr__(self) -> str:
        t = "pattern:"
        for p in self.pattern:
            t += f"\n {p}"
        return t

    @cached_property
    def regex(self) -> re.Pattern:
        t = ""
        for p in self.pattern:
            if isinstance(p, Constant):
                t += p.value
            else:
                t += f"(?P<{p.name}>\S+)"

        t = rf"^{t}$"
        return re.compile(t)


class HivePattern:
    """Hive pattern.

    Parameters
    ----------
    pattern : str
        The hive pattern string.
    values : dict, optional
        Dictionary of values for substitution, by default None.

    Attributes
    ----------
    pattern : str
        The original pattern string.
    root : str
        The root directory of the pattern. This is the beginning of the path not containing
        any pattern parameters.
    rest : str
        The remaining part of the pattern after the root.
    params : list
        List of parameters in the pattern.
    dynamic_params : list
        List of parameters without user specified values in the pattern.
    fixed_single_params : dict
        Dictionary of user specified single value parameters.
    fixed_multi_keys : dict
        Dictionary of user specified multi value parameters.
    parts : list
        List of pattern parts. Each part is a Pattern object representing a part of the path.
    """

    def __init__(self, pattern: str, values: Optional[Dict[str, TypingAny]] = None) -> None:
        self.pattern = pattern
        values = values or {}
        values = dict(values)

        # substitute single values into the pattern
        pattern = Pattern(pattern)
        self.fixed_single_params = {}
        for k in list(values.keys()):
            v = values[k]
            LOG.debug(f" {k=} {v=}")
            if isinstance(v, (list, tuple)):
                if len(v) == 1:
                    self.fixed_single_params[k] = values.pop(k)[0]
            elif v:
                self.fixed_single_params[k] = values.pop(k)

        pattern = pattern._subpattern(self.fixed_single_params)

        # analyze path structure and turn each file path part into a
        # pattern
        path = Path(pattern)
        self.root = ""
        self.rest = ""
        path_parts = path.parts
        LOG.debug(f"{pattern=} {path_parts=}")

        parts = [Pattern(x) for x in path_parts]
        self.parts = []
        for i, part in enumerate(parts):
            if part.is_constant():
                self.root = os.path.join(self.root, part.pattern[0].value)
            else:
                self.rest = os.path.join(*path_parts[i:])
                self.parts = parts[i:]
                break

        self.params = list(self.fixed_single_params.keys())
        self.dynamic_params = []
        self.fixed_multi_params = {}
        for p in self.parts:
            for v in p.variables:
                if v.name is not self.params:
                    try:
                        s = v.substitute_many(values)
                        self.fixed_multi_params[v.name] = s
                    except ValueError:
                        self.dynamic_params.append(v.name)
                    self.params.append(v.name)

        assert len(self.fixed_multi_params) == len(values), f"{len(self.fixed_multi_params)} != {len(values)}"

        LOG.debug(f"root={self.root}")
        LOG.debug(f"rest={self.rest}")
        LOG.debug(f"params={self.params}")
        LOG.debug(f"dynamic_params={self.dynamic_params}")
        LOG.debug(f"fixed_single_params={self.fixed_single_params}")
        LOG.debug(f"fixed_multi_params={self.fixed_multi_params}")
        for p in self.parts:
            LOG.debug(f" {p=}")
            LOG.debug(f"   re={p.regex}")

    def scan(self, *args: Dict[str, TypingAny], **kwargs: TypingAny) -> List[str]:
        """Scan the file system for files matching the pattern.

        Parameters
        ----------
        args : tuple of dicts
            Positional dictionaries of parameters for scanning.
        kwargs : dict
            Additional keyword arguments for scanning.

        Returns
        -------
        list
            List of file paths matching the pattern.

        Raises
        ------
        ValueError
            If a parameter is not valid for the pattern.
        """
        params_in = {}
        for a in args:
            params_in.update(a)
        params_in.update(kwargs)

        for k in list(params_in.keys()):
            if isinstance(params_in[k], tuple):
                params_in[k] = list(params_in[k])
            elif not isinstance(params_in[k], list):
                params_in[k] = [params_in[k]]

        # determine param values to use
        params = {}
        for k, v in params_in.items():
            if k in self.fixed_single_params:
                if self.fixed_single_params[k] not in v:
                    return []

            elif k in self.fixed_multi_params:
                v1 = [v1 for v1 in v if v1 in self.fixed_multi_params[k]]
                if v1:
                    params[k] = v1
                else:
                    return []
            elif k in self.dynamic_params:
                params[k] = v
            else:
                raise ValueError(f"Invalid key '{k}' not in pattern")

        for k in params:
            params[k] = set([str(x) for x in params[k]])

        root_num = len(Path(self.root).parts)
        last = len(self.parts) - 1
        res = []

        # walk the file system
        for root, dirs, files in os.walk(self.root):
            # LOG.debug(f"walk: {root=}")
            index = len(Path(root).parts) - root_num
            # LOG.debug(f"{index=} {last=}")
            part = self.parts[index]

            # intermediate level
            if index != last:
                exclude = []
                for d in dirs:
                    g = self.collect(d, part, params)
                    if g is None:
                        exclude.append(d)
                # LOG.debug(f"   {exclude=}")
                if exclude:
                    dirs[:] = [d for d in dirs if d not in exclude]
                    continue

            # last level (collection)
            else:
                for file in files:
                    # LOG.debug(f"  {file=}")
                    d = self.collect(file, part, params)
                    if d:
                        res.append(os.path.join(root, file))
                        # LOG.debug("   match")

        return res

    def collect(self, file: str, part: Pattern, params: Dict[str, set[str]]) -> Optional[Dict[str, str]]:
        # LOG.debug(f"  match={file}")
        m = part.regex.match(file)
        if m:
            if part.is_constant():
                return {}

            group = m.groupdict()
            if len(group) == len(part.variables):
                for k, v in group.items():
                    if k in params:
                        # LOG.debug(f"  {k=} {v=} {params[k]=}")
                        if v not in params[k]:
                            return None
                return group
        return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from earthkit.data.readers import Reader
from earthkit.data.utils.bbox import BoundingBox


class ShapeFileReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)
        self.__gdf = None
        assert path.endswith(".shp")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path})"

    def _repr_html_(self):
        html_repr = (
            f"<h4>{self.__class__.__name__}(represented as a geopandas object):</h3>"
            f"{self.to_pandas()._repr_html_()}"
        )

        return html_repr

    @property
    def _gdf(self):
        if self.__gdf is None:
            self.__gdf = self.to_pandas()
        return self.__gdf

    def __len__(self):
        return len(self._gdf)

    def __getitem__(self, n):
        return self._gdf.iloc[n]

    def mutate_source(self):
        # A ShapeFileReader is a source itself
        return self

    def bounding_box(self):
        return BoundingBox.from_geopandas(self._gdf)

    def ls(self, **kwargs):
        return self._gdf

    def head(self, **kwargs):
        return self._gdf.head(**kwargs)

    def tail(self, **kwargs):
        return self._gdf.tail(**kwargs)

    def describe(self, **kwargs):
        return self._gdf

    def to_numpy(self, flatten=False, **kwargs):
        v = self.to_pandas().to_numpy(**kwargs)
        if flatten:
            import math

            v = v.reshape(
                math.prod(v.shape),
            )
        return v

    def to_pandas(self, **kwargs):
        if not kwargs and self.__gdf is not None:
            return self.__gdf
        else:
            try:
                import geopandas as gpd
            except ImportError:
                raise ImportError("shapefile handling requires 'geopandas' to be installed")

            import numpy as np

            geo_df = gpd.read_file(self.path, **kwargs)
            return geo_df.set_index(np.arange(len(geo_df)))

    def to_geopandas(self, **kwargs):
        # TODO: handle multiple paths
        return self.to_pandas(**kwargs)

    def to_xarray(self, **kwargs):
        return self.to_pandas(**kwargs).to_xarray()


SHAPE_EXT = ".shp"
MANDATORY = (".shp", ".shx", ".dbf")
NON_MANDATORY = (".sbn", ".sbx", ".shp.xml", ".prj", ".CPG")
DOUBLE_DOT_EXT = tuple([e for e in NON_MANDATORY if e.count(".") == 2])


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    root, extension = os.path.splitext(path)
    for e in DOUBLE_DOT_EXT:
        if path.endswith(e):
            root = path[: -len(e)]
            extension = e

    path = root + ".shp"

    # a shapefile consists of multiple files, but we only create
    # a reader for the .shp file
    if extension == SHAPE_EXT:
        if all(os.path.exists(root + e) for e in MANDATORY):
            return ShapeFileReader(source, path)
    else:
        if extension in MANDATORY or extension in NON_MANDATORY:
            from .unknown import UnknownReader

            return UnknownReader(source, "", skip_warning=True)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import logging
import sys
from functools import cached_property

try:
    import cdsapi
except ImportError:
    raise ImportError("CDS access requires 'cdsapi' to be installed")

import yaml

from earthkit.data.core.thread import SoftThreadPool
from earthkit.data.decorators import normalize
from earthkit.data.utils import ensure_iterable
from earthkit.data.utils.progbar import tqdm

from .file import FileSource
from .prompt import APIKeyPrompt

if sys.version_info >= (3, 12):
    from itertools import batched
else:

    def batched(iterable, n):
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


LOG = logging.getLogger(__name__)


class CDSAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://cds.climate.copernicus.eu/"
    retrieve_api_key_url = "https://cds.climate.copernicus.eu/how-to-api"

    prompts = [
        dict(
            name="url",
            default="https://cds.climate.copernicus.eu/api",
            title="API url",
            validate=r"http.?://.*",
        ),
        dict(
            name="key",
            example="abcdef01-0000-1111-2222-0123456789ab",
            title="API key",
            hidden=True,
            validate=r"[\-0-9a-f]+",
        ),
    ]

    rcfile = "~/.cdsapirc"
    rcfile_env = "CDSAPI_RC"
    config_env = ("CDSAPI_URL", "CDSAPI_KEY")

    def save(self, input, file):
        yaml.dump(input, file, default_flow_style=False)

    def load(self, file):
        return yaml.safe_load(file.read())


def client(use_prompt):
    if use_prompt:
        prompt = CDSAPIKeyPrompt()
        prompt.check()

        try:
            return cdsapi.Client()
        except Exception as e:
            if ".cdsapirc" in str(e) or not prompt.has_config_env():
                LOG.warning(e)
                LOG.exception(f"Could not load cds api client. {e}")
                prompt.ask_user_and_save()
                return cdsapi.Client()

            raise
    else:
        return cdsapi.Client()


EXTENSIONS = {
    "grib": ".grib",
    "netcdf": ".nc",
}


class CdsRetriever(FileSource):
    sphinxdoc = """
    CdsRetriever
    """

    def __init__(self, dataset, *args, prompt=True, **kwargs):
        super().__init__()

        self.prompt = prompt

        assert isinstance(dataset, str)
        if args and kwargs:
            raise TypeError("CdsRetriever: cannot specify request using both args and kwargs")

        if not args:
            args = (kwargs,)
        assert all(isinstance(request, dict) for request in args)
        self._args = args

        self.client()  # Trigger password prompt before threading

        nthreads = min(self.config("number-of-download-threads"), len(self.requests))

        if nthreads < 2:
            self.path = [self._retrieve(dataset, r) for r in self.requests]
        else:
            with SoftThreadPool(nthreads=nthreads) as pool:
                futures = [pool.submit(self._retrieve, dataset, r) for r in self.requests]

                iterator = (f.result() for f in futures)
                self.path = list(tqdm(iterator, leave=True, total=len(self.requests)))

    def _retrieve(self, dataset, request):
        def retrieve(target, args):
            cds_result = self.client().retrieve(args[0], args[1])
            self.source_filename = cds_result.location.split("/")[-1]
            cds_result.download(target=target)

        return_object = self.cache_file(
            retrieve,
            (dataset, request),
            extension=EXTENSIONS.get(request.get("format"), ".cache"),
        )
        return return_object

    @staticmethod
    @normalize("area", "bounding-box(list)")
    def _normalize_request(**kwargs):
        return kwargs

    @cached_property
    def requests(self):
        requests = []
        for arg in self._args:
            request = self._normalize_request(**arg)
            split_on = request.pop("split_on", None)
            if split_on is None:
                requests.append(request)
                continue

            if not isinstance(split_on, dict):
                split_on = {k: 1 for k in ensure_iterable(split_on)}
            for values in itertools.product(
                *[batched(ensure_iterable(request[k]), v) for k, v in split_on.items()]
            ):
                subrequest = dict(zip(split_on, values))
                requests.append({**request, **subrequest})
        return requests

    def client(self):
        return client(self.prompt)


source = CdsRetriever
